#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re


def getAvailableShips(session):
    """Function that returns the total number of free (available) ships
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    ships : int
        number of currently available ships
    """
    html = session.get()
    return int(re.search(r'GlobalMenu_freeTransporters">(\d+)<', html).group(1))


def getTotalShips(session):
    """Function that returns the total number of ships, regardless of if they're available or not
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    ships : int
        total number of ships the player has
    """
    html = session.get()
    return int(re.search(r'maxTransporters">(\d+)<', html).group(1))



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
import gettext
import json
import sys
import requests
from ikabot.helpers.pedirInfo import read
from ikabot.helpers.gui import *
from ikabot.config import *

t = gettext.translation('insertCookies', localedir, languages=languages, fallback=True)
_ = t.gettext


def importExportCookie(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    banner()
    try:
        print('Do you want to import or export the cookie?')
        print('(0) Exit')
        print('(1) Import')
        print('(2) Export')
        action = read(min=0, max=2)
        if action == 1:
            importCookie(session)
        elif action == 2:
            exportCookie(session)

        event.set()
    except KeyboardInterrupt:
        event.set()
        return


def importCookie(session):
    banner()
    print('{}⚠️ INSERTING AN INVALID COOKIE WILL LOG YOU OUT OF YOUR OTHER SESSIONS ⚠️{}\n\n'.format(bcolors.WARNING, bcolors.ENDC))
    print('Go ahead and export the cookie from another ikabot instance now and then')
    print('type your "ikariam" cookie below:')
    newcookie = read()
    newcookie = newcookie.strip()
    newcookie = newcookie.replace('ikariam=', '')
    cookies = session.getSessionData()['cookies']
    cookies['ikariam'] = newcookie
    if session.host in session.s.cookies._cookies:
        session.s.cookies.set('ikariam', newcookie, domain=session.host, path='/')
    else:
        session.s.cookies.set('ikariam', newcookie, domain='', path='/')

    html = session.s.get(session.urlBase).text

    if session.isExpired(html):
        print('{}Failure!{} All your other sessions have just been invalidated!'.format(bcolors.RED, bcolors.ENDC))
        enter()
    else:
        print('{}Success!{} This ikabot session will now use the cookie you provided'.format(bcolors.GREEN, bcolors.ENDC))
        sessionData = session.getSessionData()
        sessionData['cookies']['ikariam'] = newcookie
        session.setSessionData(sessionData)
        enter()
    session.get()


def exportCookie(session):
    banner()
    session.get()  # get valid cookie in case user has logged the bot out before running this feature
    ikariam = session.getSessionData()['cookies']['ikariam']
    print('Use this cookie to synchronise two ikabot instances on 2 different machines\n\n')
    print('ikariam='+ikariam+'\n\n')

    cookie = json.dumps({"ikariam": ikariam})  # get ikariam cookie, only this cookie is invalidated when the bot logs the user out.
    cookies_js = 'cookies={};i=0;for(let cookie in cookies){{document.cookie=Object.keys(cookies)[i]+\"=\"+cookies[cookie];i++}}'.format(cookie)
    print("""To prevent ikabot from logging you out while playing Ikariam do the following:
    1. Be on the "Your session has expired" screen
    2. Open Chrome javascript console by pressing CTRL + SHIFT + J
    3. Copy the text below, paste it into the console and press enter
    4. Press F5
    """)
    print(cookies_js)
    enter()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import gettext
import sys
import requests
import re
import traceback
import time
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.gui import *
from ikabot.config import *
from ikabot.helpers.getJson import *
from ikabot.helpers.botComm import *
from ikabot.helpers.varios import wait
from ikabot.helpers.process import run


t = gettext.translation('buyResources', localedir, languages=languages, fallback=True)
_ = t.gettext


def autoPirate(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    banner()
    try:
        if not isWindows:
            path = run('which nslookup')
            is_installed = re.search(r'/.*?/nslookup', path) is not None
            if is_installed is False:
                print('you must first install nslookup')
                enter()
                event.set()
                return

        print('{}⚠️ USING THIS FEATURE WILL EXPOSE YOUR IP ADDRESS TO A THIRD PARTY FOR CAPTCHA SOLVING ⚠️{}\n\n'.format(bcolors.WARNING, bcolors.ENDC))
        print('How many pirate missions should I do? (min = 1)')
        pirateCount = read(min=1, digit=True)
        print("""Which pirate mission should I do?
    (1) 2m 30s
    (2) 7m 30s
    (3) 15m
    (4) 30m
    (5) 1h
    (6) 2h
    (7) 4h
    (8) 8h
    (9) 16h
    """)
        pirateMissionChoice = read(min=1, max=9, digit=True)
        print('Do you want me to automatically convert capture points to crew strength? (Y|N)')
        autoConvert = read(values=['y', 'Y', 'n', 'N'])
        if autoConvert.lower() == 'y':
            print('How many points should I convert every time I do a mission? (Type "all" to convert all points at once)')
            convertPerMission = read(min=0, additionalValues=['all'], digit=True)
        print('Enter a maximum additional random waiting time between missions in seconds. (min = 0)')
        maxRandomWaitingTime = read(min=0, digit=True)
        piracyCities = getPiracyCities(session, pirateMissionChoice)
        if piracyCities == []:
            print('You do not have any city with a pirate fortress capable of executing this mission!')
            enter()
            event.set()
            return

        print('YAAAAAR!')  # get data for options such as auto-convert to crew strength, time intervals, number of piracy attempts... ^^
        enter()
    except KeyboardInterrupt:
        event.set()
        return
    event.set()
    try:
        while (pirateCount > 0):
            pirateCount -= 1
            piracyCities = getPiracyCities(session, pirateMissionChoice)  # this is done again inside the loop in case the user destroys / creates another pirate fortress while this module is running
            if piracyCities == []:
                raise Exception('No city with pirate fortress capable of executing selected mission')
            html = session.post(city_url + str(piracyCities[0]['id']))  # this is needed because for some reason you need to look at the town where you are sending a request from in the line below, before you send that request
            if '"showPirateFortressShip":0' in html:  # this is in case the user has manually run a capture run, in that case, there is no need to wait 150secs instead we can check every 5
                url = 'view=pirateFortress&cityId={}&position=17&backgroundView=city&currentCityId={}&actionRequest={}&ajax=1'.format(piracyCities[0]['id'], piracyCities[0]['id'], actionRequest)
                html = session.post(url)
                wait(getCurrentMissionWaitingTime(html), maxRandomWaitingTime)
                pirateCount += 1  # don't count this as an iteration of the loop
                continue

            url = 'action=PiracyScreen&function=capture&buildingLevel={0}&view=pirateFortress&cityId={1}&position=17&activeTab=tabBootyQuest&backgroundView=city&currentCityId={1}&templateView=pirateFortress&actionRequest={2}&ajax=1'.format(piracyMissionToBuildingLevel[pirateMissionChoice], piracyCities[0]['id'], actionRequest)
            html = session.post(url)

            if 'function=createCaptcha' in html:
                try:
                    for i in range(20):
                        if i == 19:
                            raise Exception("Failed to resolve captcha too many times")
                        picture = session.get('action=Options&function=createCaptcha', fullResponse=True).content
                        captcha = resolveCaptcha(session, picture)
                        if captcha == 'Error':
                            continue
                        session.post(city_url + str(piracyCities[0]['id']))
                        params = {'action': 'PiracyScreen', 'function': 'capture', 'cityId': piracyCities[0]['id'], 'position': '17', 'captchaNeeded': '1', 'buildingLevel': str(piracyMissionToBuildingLevel[pirateMissionChoice]), 'captcha': captcha, 'activeTab': 'tabBootyQuest', 'backgroundView': 'city', 'currentCityId': piracyCities[0]['id'], 'templateView': 'pirateFortress', 'actionRequest': actionRequest, 'ajax': '1'}
                        html = session.post(payloadPost=params, noIndex=True)
                        if '"showPirateFortressShip":1' in html:  # if this is true, then the crew is still in the town, that means that the request didn't succeed
                            continue
                        break
                except Exception:
                    info = ''
                    msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
                    sendToBot(session, msg)
                    break
            if autoConvert.lower() == 'y':
                convertCapturePoints(session, piracyCities, convertPerMission)
            wait(piracyMissionWaitingTime[pirateMissionChoice], maxRandomWaitingTime)

    except Exception:
        event.set()
        return


def resolveCaptcha(session, picture):
    session_data = session.getSessionData()
    if 'decaptcha' not in session_data or session_data['decaptcha']['name'] == 'default':
        text = run('nslookup -q=txt ikagod.twilightparadox.com ns2.afraid.org')
        parts = text.split('"')
        if len(parts) < 2:
            # the DNS output is not well formed
            return 'Error'
        address = parts[1]

        files = {'upload_file': picture}
        captcha = requests.post('http://{0}'.format(address), files=files).text
        return captcha
    elif session_data['decaptcha']['name'] == 'custom':
        files = {'upload_file': picture}
        captcha = requests.post('{0}'.format(session_data['decaptcha']['endpoint']), files=files).text
        return captcha
    elif session_data['decaptcha']['name'] == '9kw.eu':
        credits = requests.get("https://www.9kw.eu/index.cgi?action=usercaptchaguthaben&apikey={}".format(session_data['decaptcha']['relevant_data']['apiKey'])).text
        if int(credits) < 10:
            raise Exception('You do not have enough 9kw.eu credits!')
        captcha_id = requests.post("https://www.9kw.eu/index.cgi?action=usercaptchaupload&apikey={}".format(session_data['decaptcha']['relevant_data']['apiKey']), headers={'Content-Type': 'multipart/form-data'}, files={'file-upload-01': picture}).text
        while True:
            captcha_result = requests.get("https://www.9kw.eu/index.cgi?action=usercaptchacorrectdata&id={}&apikey={}".format(captcha_id, session_data['decaptcha']['relevant_data']['apiKey'])).text
            if captcha_result != '':
                return captcha_result.upper()
            wait(5)
    elif session_data['decaptcha']['name'] == 'telegram':
        sendToBot(session, 'Please solve the captcha', Photo=picture)
        captcha_time = time.time()
        while(True):
            response = getUserResponse(session, fullResponse=True)
            if len(response) == 0:
                time.sleep(5)
                continue
            response = response[-1]
            if response['date'] > captcha_time:
                return response['text']
            time.sleep(5)


def getPiracyCities(session, pirateMissionChoice):
    """Gets all user's cities which have a pirate fortress in them
    Parameters
    ----------
    session : ikabot.web.session.Session

    Returns
    -------
    piracyCities : list[dict]
    """
    cities_ids = getIdsOfCities(session)[0]
    piracyCities = []
    for city_id in cities_ids:
        html = session.get(city_url + city_id)
        city = getCity(html)
        for pos, building in enumerate(city['position']):
            if building['building'] == 'pirateFortress' and building['level'] >= piracyMissionToBuildingLevel[pirateMissionChoice]:
                piracyCities.append(city)
                break
    return piracyCities


def convertCapturePoints(session, piracyCities, convertPerMission):
    """Converts all the users capture points into crew strength
    Parameters
    ----------
    session : ikabot.web.session.Session
    piracyCities: a list containing all cities which have a pirate fortress
    """
    html = session.get('view=pirateFortress&activeTab=tabCrew&cityId={0}&position=17&backgroundView=city&currentCityId={0}&templateView=pirateFortress'.format(piracyCities[0]['id']))
    rta = re.search(r'\\"capturePoints\\":\\"(\d+)\\"', html)
    capturePoints = int(rta.group(1))
    if convertPerMission == 'all':
        convertPerMission = capturePoints
    if 'conversionProgressBar' in html:  # if a conversion is still in progress
        return
    data = {'action': 'PiracyScreen', 'function': 'convert', 'view': 'pirateFortress', 'cityId': piracyCities[0]['id'], 'islandId': piracyCities[0]['islandId'], 'activeTab': 'tabCrew', 'crewPoints': str(int(convertPerMission/10)), 'position': '17', 'backgroundView': 'city', 'currentCityId': piracyCities[0]['id'], 'templateView': 'pirateFortress', 'actionRequest': actionRequest, 'ajax': '1'}
    html = session.post(payloadPost=data, noIndex=True)


def getCurrentMissionWaitingTime(html):
    match = re.search(r'missionProgressTime\\\\">(.*?)<\\\\\/div>', html)
    if match is None:
        return 0
    else:
        time_string = match.group(1)
        hours = re.search(r'(\d+)h', time_string)
        if hours is None:
            hours = 0
        else:
            hours = int(hours.group(1)) * 3600
        minutes = re.search(r'(\d+)m', time_string)
        if minutes is None:
            minutes = 0
        else:
            minutes = int(minutes.group(1)) * 60
        seconds = re.search(r'(\d+)s', time_string)
        if seconds is None:
            seconds = 0
        else:
            seconds = int(seconds.group(1)) * 1
        return hours + minutes + seconds



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import signal
import gettext
from ikabot.config import *
from ikabot.helpers.botComm import *

t = gettext.translation('signals', localedir, languages=languages, fallback=True)
_ = t.gettext


def do_nothing(signal, frame):
    pass


def deactivate_sigint():
    signal.signal(signal.SIGINT, do_nothing)


def create_handler(s):
    def _handler(signum, frame):
        raise Exception(_('Signal number {:d} received').format(signum))
    return _handler


def setSignalsHandlers(s):
    signals = [signal.SIGINT, signal.SIGTERM]  # signal.SIGQUIT replaced with signal.SIGINT for compatibility
    for sgn in signals:
        signal.signal(sgn, create_handler(s))


def setInfoSignal(session, info):  # send process info to bot
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    info : str
    """
    info = _('information of the process {}:\n{}').format(os.getpid(), info)

    def _sendInfo(signum, frame):
        sendToBot(session, info)
    signal.signal(signal.SIGABRT, _sendInfo)  # kill -SIGUSR1 pid, SIGUSR1 replaced with SIGABRT for compatibility



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import gettext
import traceback
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.varios import *
from ikabot.helpers.botComm import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.getJson import getCity
from ikabot.helpers.signals import setInfoSignal

t = gettext.translation('activateMiracle', localedir, languages=languages, fallback=True)
_ = t.gettext


def obtainMiraclesAvailable(session):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session

    Returns
    -------
    islands: list[dict]
    """
    idsIslands = getIslandsIds(session)
    islands = []
    for idIsland in idsIslands:
        html = session.get(island_url + idIsland)
        island = getIsland(html)
        island['activable'] = False
        islands.append(island)

    ids, cities = getIdsOfCities(session)
    for city_id in cities:
        city = cities[city_id]
        # get the wonder for this city
        wonder = [island['wonder'] for island in islands if city['coords'] == '[{}:{}] '.format(island['x'], island['y'])][0]
        # if the wonder is not new, continue
        if wonder in [island['wonder'] for island in islands if island['activable']]:
            continue

        html = session.get(city_url + str(city['id']))
        city = getCity(html)

        # make sure that the city has a temple
        for i in range(len(city['position'])):
            if city['position'][i]['building'] == 'temple':
                city['pos'] = str(i)
                break
        else:
            continue

        # get wonder information
        params = {"view": "temple", "cityId": city['id'], "position": city['pos'], "backgroundView": "city", "currentCityId": city['id'], "actionRequest": actionRequest, "ajax": "1"}
        data = session.post(params=params)
        data = json.loads(data, strict=False)
        data = data[2][1]
        available = data['js_WonderViewButton']['buttonState'] == 'enabled'
        if available is False:
            for elem in data:
                if 'countdown' in data[elem]:
                    enddate = data[elem]['countdown']['enddate']
                    currentdate = data[elem]['countdown']['currentdate']
                    break

        # set the information on the island which wonder we can activate
        for island in islands:
            if island['id'] == city['islandId']:
                island['activable'] = True
                island['ciudad'] = city
                island['available'] = available
                if available is False:
                    island['available_in'] = enddate - currentdate
                break

    # only return island which wonder we can activate
    return [island for island in islands if island['activable']]


def activateMiracleHttpCall(session, island):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    island : dict

    Returns
    -------
    json : dict
    """
    params = {'action': 'CityScreen', 'cityId': island['ciudad']['id'], 'function': 'activateWonder', 'position': island['ciudad']['pos'], 'backgroundView': 'city', 'currentCityId': island['ciudad']['id'], 'templateView': 'temple', 'actionRequest': actionRequest, 'ajax': '1'}
    response = session.post(params=params)
    return json.loads(response, strict=False)


def chooseIsland(islands):
    """
    Parameters
    ----------
    islands : list[dict]

    Returns
    -------
    island : dict
    """
    print(_('Which miracle do you want to activate?'))
    i = 0
    print(_('(0) Exit'))
    for island in islands:
        i += 1
        if island['available']:
            print('({:d}) {}'.format(i, island['wonderName']))
        else:
            print(_('({:d}) {} (available in: {})').format(i, island['wonderName'], daysHoursMinutes(island['available_in'])))

    index = read(min=0, max=i)
    if index == 0:
        return None
    island = islands[index - 1]
    return island


def activateMiracle(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        islands = obtainMiraclesAvailable(session)
        if islands == []:
            print(_('There are no miracles available.'))
            enter()
            event.set()
            return

        island = chooseIsland(islands)
        if island is None:
            event.set()
            return

        if island['available']:
            print(_('\nThe miracle {} will be activated').format(island['wonderName']))
            print(_('Proceed? [Y/n]'))
            activate_miracle_input = read(values=['y', 'Y', 'n', 'N', ''])
            if activate_miracle_input.lower() == 'n':
                event.set()
                return

            miracle_activation_result = activateMiracleHttpCall(session, island)

            if miracle_activation_result[1][1][0] == 'error':
                print(_('The miracle {} could not be activated.').format(island['wonderName']))
                enter()
                event.set()
                return

            data = miracle_activation_result[2][1]
            for elem in data:
                if 'countdown' in data[elem]:
                    enddate = data[elem]['countdown']['enddate']
                    currentdate = data[elem]['countdown']['currentdate']
                    break
            wait_time = enddate - currentdate

            print(_('The miracle {} was activated.').format(island['wonderName']))
            enter()
            banner()

            while True:
                print(_('Do you wish to activate it again when it is finished? [y/N]'))

                reactivate_again_input = read(values=['y', 'Y', 'n', 'N', ''])
                if reactivate_again_input.lower() != 'y':
                    event.set()
                    return

                iterations = read(msg=_('How many times?: '), digit=True, min=0)

                if iterations == 0:
                    event.set()
                    return

                duration = wait_time * iterations

                print(_('It will finish in:{}').format(daysHoursMinutes(duration)))

                print(_('Proceed? [Y/n]'))
                reactivate_again_input = read(values=['y', 'Y', 'n', 'N', ''])
                if reactivate_again_input.lower() == 'n':
                    banner()
                    continue
                break
        else:
            print(_('\nThe miracle {} will be activated in {}').format(island['wonderName'], daysHoursMinutes(island['available_in'])))
            print(_('Proceed? [Y/n]'))
            user_confirm = read(values=['y', 'Y', 'n', 'N', ''])
            if user_confirm.lower() == 'n':
                event.set()
                return
            wait_time = island['available_in']
            iterations = 1

            print(_('\nThe mirable will be activated.'))
            enter()
            banner()

            while True:
                print(_('Do you wish to activate it again when it is finished? [y/N]'))

                reactivate_again_input = read(values=['y', 'Y', 'n', 'N', ''])
                again = reactivate_again_input.lower() == 'y'
                if again is True:
                    try:
                        iterations = read(msg=_('How many times?: '), digit=True, min=0)
                    except KeyboardInterrupt:
                        iterations = 1
                        break

                    if iterations == 0:
                        iterations = 1
                        break

                    iterations += 1
                    duration = wait_time * iterations
                    print(_('It is not possible to calculate the time of finalization. (at least: {})').format(daysHoursMinutes(duration)))
                    print(_('Proceed? [Y/n]'))

                    try:
                        activate_input = read(values=['y', 'Y', 'n', 'N', ''])
                    except KeyboardInterrupt:
                        iterations = 1
                        break

                    if activate_input.lower() == 'n':
                        iterations = 1
                        banner()
                        continue
                break
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nI activate the miracle {} {:d} times\n').format(island['wonderName'], iterations)
    setInfoSignal(session, info)
    try:
        do_it(session, island, iterations)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def wait_for_miracle(session, island):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    island : dict
    """
    while True:
        params = {"view": "temple", "cityId": island['ciudad']['id'], "position": island['ciudad']['pos'], "backgroundView": "city", "currentCityId": island['ciudad']['id'], "actionRequest": actionRequest, "ajax": "1"}
        temple_response = session.post(params=params)
        temple_response = json.loads(temple_response, strict=False)
        temple_response = temple_response[2][1]

        for elem in temple_response:
            if 'countdown' in temple_response[elem]:
                enddate = temple_response[elem]['countdown']['enddate']
                currentdate = temple_response[elem]['countdown']['currentdate']
                wait_time = enddate - currentdate
                break
        else:
            available = temple_response['js_WonderViewButton']['buttonState'] == 'enabled'
            if available:
                return
            else:
                wait_time = 60

        msg = _('I wait {:d} seconds to activate the miracle {}').format(wait_time, island['wonderName'])
        sendToBotDebug(session, msg, debugON_activateMiracle)
        wait(wait_time + 5)


def do_it(session, island, iterations):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    island : dict
    iterations : int
    """
    for i in range(iterations):

        wait_for_miracle(session, island)

        response = activateMiracleHttpCall(session, island)

        if response[1][1][0] == 'error':
            msg = _('The miracle {} could not be activated.').format(island['wonderName'])
            sendToBot(session, msg)
            return

        msg = _('Miracle {} successfully activated').format(island['wonderName'])
        sendToBotDebug(session, msg, debugON_activateMiracle)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import json
import psutil
import subprocess
from ikabot.config import *
from ikabot.helpers.signals import deactivate_sigint


def set_child_mode(session):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    """
    session.padre = False
    deactivate_sigint()


def run(command):
    ret = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.read()
    try:
        return ret.decode('utf-8').strip()
    except Exception:
        return ret


def updateProcessList(session, programprocesslist=[]):
    """This function will return data about all the active ikabot processes. If it is passed the ``programprocesslist`` argument, it will write new processes from that list to the .ikabot file
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    programprocesslist : list[dict]
        a list of dictionaries containing relevant data about a running ikabot process ('pid', 'proxies' and 'action')

    Returns
    -------
    runningIkabotProcessList : list[dict]
        a list of dictionaries containing relevant data about a running ikabot process ('pid', 'proxies' and 'action')
    """
    # read from file
    sessionData = session.getSessionData()
    try:
        fileList = sessionData['processList']
    except KeyError:
        fileList = []

    # check it's still running
    runningIkabotProcessList = []
    ika_process = psutil.Process(pid=os.getpid()).name()
    for process in fileList:
        try:
            proc = psutil.Process(pid=process['pid'])
        except psutil.NoSuchProcess:
            continue

        # windows doesn't support the status method
        isAlive = True if isWindows else proc.status() != 'zombie'

        if proc.name() == ika_process and isAlive:
            runningIkabotProcessList.append(process)

    # add new to the list and write to file only if it's given
    for process in programprocesslist:
        if process not in runningIkabotProcessList:
            runningIkabotProcessList.append(process)

    # write to file
    sessionData['processList'] = runningIkabotProcessList
    session.setSessionData(sessionData)

    return runningIkabotProcessList



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import gettext
import traceback
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.botComm import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.planRoutes import executeRoutes
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.getJson import getCity
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.varios import addThousandSeparator
from ikabot.helpers.resources import *

t = gettext.translation('sendResources', localedir, languages=languages, fallback=True)
_ = t.gettext


def sendResources(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        routes = []
        while True:

            banner()
            print(_('Origin city:'))
            try:
                cityO = chooseCity(session)
            except KeyboardInterrupt:
                if routes:
                    print(_('Send shipment? [Y/n]'))
                    rta = read(values=['y', 'Y', 'n', 'N', ''])
                    if rta.lower() != 'n':
                        break
                event.set()
                return

            banner()
            print(_('Destination city'))
            cityD = chooseCity(session, foreign=True)
            idIsland = cityD['islandId']

            if cityO['id'] == cityD['id']:
                continue

            resources_left = cityO['recursos']
            for route in routes:
                (origin_city, destination_city, __, *toSend) = route
                if origin_city['id'] == cityO['id']:
                    for i in range(len(materials_names)):
                        resources_left[i] -= toSend[i]

                # the destination city might be from another player
                if cityD['propia'] and destination_city['id'] == cityD['id']:
                    for i in range(len(materials_names)):
                        cityD['freeSpaceForResources'][i] -= toSend[i]

            banner()
            # the destination city might be from another player
            if cityD['propia']:
                msg = ''
                for i in range(len(materials_names)):
                    if resources_left[i] > cityD['freeSpaceForResources'][i]:
                        msg += '{} more {}\n'.format(addThousandSeparator(cityD['freeSpaceForResources'][i]), materials_names[i].lower())

                if len(msg) > 0:
                    print(_('You can store just:\n{}').format(msg))

            print(_('Available:'))
            for i in range(len(materials_names)):
                print('{}:{} '.format(materials_names[i], addThousandSeparator(resources_left[i])), end='')
            print('')

            print(_('Send:'))
            try:
                max_name = max([len(material) for material in materials_names])
                send = []
                for i in range(len(materials_names)):
                    material_name = materials_names[i]
                    pad = ' ' * (max_name - len(material_name))
                    val = askForValue(_('{}{}:'.format(pad, material_name)), resources_left[i])
                    send.append(val)
            except KeyboardInterrupt:
                continue
            if sum(send) == 0:
                continue

            banner()
            print(_('About to send from {} to {}').format(cityO['cityName'], cityD['cityName']))
            for i in range(len(materials_names)):
                if send[i] > 0:
                    print('{}:{} '.format(materials_names[i], addThousandSeparator(send[i])), end='')
            print('')

            print(_('Proceed? [Y/n]'))
            rta = read(values=['y', 'Y', 'n', 'N', ''])
            if rta.lower() != 'n':
                route = (cityO, cityD, idIsland, *send)
                routes.append(route)
                print(_('Create another shipment? [y/N]'))
                rta = read(values=['y', 'Y', 'n', 'N', ''])
                if rta.lower() != 'y':
                    break
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nSend resources\n')

    setInfoSignal(session, info)
    try:
        executeRoutes(session, routes)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import json
import gettext
import traceback
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.botComm import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.varios import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.varios import addThousandSeparator
from ikabot.helpers.getJson import getCity
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.resources import getAvailableResources

t = gettext.translation('trainArmy',
                        localedir,
                        languages=languages,
                        fallback=True)
_ = t.gettext


def getBuildingInfo(session, city, trainTroops):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    trainTroops : bool

    Returns
    -------
    response : dict
    """
    view = 'barracks' if trainTroops else 'shipyard'
    params = {'view': view, 'cityId': city['id'], 'position': city['pos'], 'backgroundView': 'city', 'currentCityId': city['id'], 'actionRequest': actionRequest, 'ajax': '1'}
    data = session.post(params=params)
    return json.loads(data, strict=False)


def train(session, city, trainings, trainTroops):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    trainings : list[dict]
    trainTroops : bool
    """
    templateView = 'barracks' if trainTroops else 'shipyard'
    function = 'buildUnits' if trainTroops else 'buildShips'
    payload = {'action': 'CityScreen', 'function': function, 'actionRequest': 'REQUESTID', 'cityId': city['id'], 'position': city['pos'], 'backgroundView': 'city', 'currentCityId': city['id'], 'templateView': templateView, 'ajax': '1'}
    for training in trainings:
        payload[training['unit_type_id']] = training['train']
    session.post(payloadPost=payload)


def waitForTraining(session, city, trainTroops):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    trainTroops : bool
    """
    data = getBuildingInfo(session, city, trainTroops)
    html = data[1][1][1]
    seconds = re.search(r'\'buildProgress\', (\d+),', html)
    if seconds:
        seconds = seconds.group(1)
        seconds = int(seconds) - data[0][1]['time']
        wait(seconds + 5)


def planTrainings(session, city, trainings, trainTroops):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    trainings : list[list[dict]]
    trainTroops : bool
    """
    buildingPos = city['pos']

    # trainings might be divided in multriple rounds
    while True:

        # total number of units to create
        total = sum([unit['cantidad'] for training in trainings for unit in training])
        if total == 0:
            return

        for training in trainings:
            waitForTraining(session, city, trainTroops)
            html = session.get(city_url + city['id'])
            city = getCity(html)
            city['pos'] = buildingPos

            resourcesAvailable = city['recursos'].copy()
            resourcesAvailable.append(city['ciudadanosDisp'])

            # for each unit type in training
            for unit in training:

                # calculate how many units can actually be trained based on the resources available
                unit['train'] = unit['cantidad']

                for i in range(len(materials_names_english)):
                    material_name = materials_names_english[i].lower()
                    if material_name in unit['costs']:
                        limiting = resourcesAvailable[i] // unit['costs'][material_name]
                        unit['train'] = min(unit['train'], limiting)

                if 'citizens' in unit['costs']:
                    limiting = resourcesAvailable[len(materials_names_english)] // unit['costs']['citizens']
                    unit['train'] = min(unit['train'], limiting)

                # calculate the resources that will be left
                for i in range(len(materials_names_english)):
                    material_name = materials_names_english[i].lower()
                    if material_name in unit['costs']:
                        resourcesAvailable[i] -= unit['costs'][material_name] * unit['train']

                if 'citizens' in unit['costs']:
                    resourcesAvailable[len(materials_names_english)] -= unit['costs']['citizens'] * unit['train']

                unit['cantidad'] -= unit['train']

            # amount of units that will be trained
            total = sum([unit['train'] for unit in training])
            if total == 0:
                msg = _('It was not possible to finish the training due to lack of resources.')
                sendToBot(session, msg)
                return

            train(session, city, training, trainTroops)


def generateArmyData(units_info):
    """
    Parameters
    ----------
    units_info : dict

    Returns
    -------
    units : list[dict]
    """
    i = 1
    units = []
    while 'js_barracksSlider{:d}'.format(i) in units_info:
        # {"identifier":"phalanx","unit_type_id":303,"costs":{"citizens":1,"wood":27,"sulfur":30,"upkeep":3,"completiontime":71.169695412658},"local_name":"Hoplita"}
        info = units_info['js_barracksSlider{:d}'.format(i)]['slider']['control_data']
        info = json.loads(info, strict=False)
        units.append(info)
        i += 1
    return units


def trainArmy(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        print(_('Do you want to train troops (1) or ships (2)?'))
        rta = read(min=1, max=2)
        trainTroops = rta == 1
        banner()

        if trainTroops:
            print(_('In what city do you want to train the troops?'))
        else:
            print(_('In what city do you want to train the fleet?'))
        city = chooseCity(session)
        banner()

        lookfor = 'barracks' if trainTroops else 'shipyard'
        for i in range(len(city['position'])):
            if city['position'][i]['building'] == lookfor:
                city['pos'] = str(i)
                break
        else:
            if trainTroops:
                print(_('Barracks not built.'))
            else:
                print(_('Shipyard not built.'))
            enter()
            event.set()
            return

        data = getBuildingInfo(session, city, trainTroops)

        units_info = data[2][1]
        units = generateArmyData(units_info)

        maxSize = max([len(unit['local_name']) for unit in units])

        tranings = []
        while True:
            units = generateArmyData(units_info)
            print(_('Train:'))
            for unit in units:
                pad = ' ' * (maxSize - len(unit['local_name']))
                amount = read(msg='{}{}:'.format(pad, unit['local_name']), min=0, empty=True)
                if amount == '':
                    amount = 0
                unit['cantidad'] = amount

            # calculate costs
            cost = [0] * (len(materials_names_english) + 3)
            for unit in units:
                for i in range(len(materials_names_english)):
                    material_name = materials_names_english[i].lower()
                    if material_name in unit['costs']:
                        cost[i] += unit['costs'][material_name] * unit['cantidad']

                if 'citizens' in unit['costs']:
                    cost[len(materials_names_english)+0] += unit['costs']['citizens'] * unit['cantidad']
                if 'upkeep' in unit['costs']:
                    cost[len(materials_names_english)+1] += unit['costs']['upkeep'] * unit['cantidad']
                if 'completiontime' in unit['costs']:
                    cost[len(materials_names_english)+2] += unit['costs']['completiontime'] * unit['cantidad']

            print(_('\nTotal cost:'))
            for i in range(len(materials_names_english)):
                if cost[i] > 0:
                    print('{}: {}'.format(materials_names_english[i], addThousandSeparator(cost[i])))
            if cost[len(materials_names_english)+0] > 0:
                print(_('Citizens: {}').format(addThousandSeparator(cost[len(materials_names_english)+0])))
            if cost[len(materials_names_english)+1] > 0:
                print(_('Maintenance: {}').format(addThousandSeparator(cost[len(materials_names_english)+1])))
            if cost[len(materials_names_english)+2] > 0:
                print(_('Duration: {}').format(daysHoursMinutes(int(cost[len(materials_names_english)+2]))))

            print(_('\nProceed? [Y/n]'))
            rta = read(values=['y', 'Y', 'n', 'N', ''])
            if rta.lower() == 'n':
                event.set()
                return

            tranings.append(units)

            if trainTroops:
                print(_('\nDo you want to train more troops when you finish? [y/N]'))
            else:
                print(_('\nDo you want to train more fleets when you finish? [y/N]'))
            rta = read(values=['y', 'Y', 'n', 'N', ''])
            if rta.lower() == 'y':
                banner()
                continue
            else:
                break

        # calculate if the city has enough resources
        resourcesAvailable = city['recursos'].copy()
        resourcesAvailable.append(city['ciudadanosDisp'])

        for training in tranings:
            for unit in training:

                for i in range(len(materials_names_english)):
                    material_name = materials_names_english[i].lower()
                    if material_name in unit['costs']:
                        resourcesAvailable[i] -= unit['costs'][material_name] * unit['cantidad']

                if 'citizens' in unit['costs']:
                    resourcesAvailable[len(materials_names_english)] -= unit['costs']['citizens'] * unit['cantidad']

        not_enough = [elem for elem in resourcesAvailable if elem < 0] != []

        if not_enough:
            print(_('\nThere are not enough resources:'))
            for i in range(len(materials_names_english)):
                if resourcesAvailable[i] < 0:
                    print('{}:{}'.format(materials_names[i], addThousandSeparator(resourcesAvailable[i]*-1)))

            if resourcesAvailable[len(materials_names_english)] < 0:
                print(_('Citizens:{}').format(addThousandSeparator(resourcesAvailable[len(materials_names_english)]*-1)))

            print(_('\nProceed anyway? [Y/n]'))
            rta = read(values=['y', 'Y', 'n', 'N', ''])
            if rta.lower() == 'n':
                event.set()
                return

        if trainTroops:
            print(_('\nThe selected troops will be trained.'))
        else:
            print(_('\nThe selected fleet will be trained.'))
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    if trainTroops:
        info = _('\nI train troops in {}\n').format(city['cityName'])
    else:
        info = _('\nI train fleets in {}\n').format(city['cityName'])
    setInfoSignal(session, info)
    try:
        planTrainings(session, city, tranings, trainTroops)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import time
import traceback
import gettext
from decimal import *
from ikabot.config import *
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import getIdsOfCities
from ikabot.helpers.varios import daysHoursMinutes
from ikabot.helpers.getJson import getCity
from ikabot.helpers.resources import *
from ikabot.helpers.botComm import *

t = gettext.translation('alertLowWine', localedir, languages=languages, fallback=True)
_ = t.gettext

getcontext().prec = 30


def alertLowWine(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        if checkTelegramData(session) is False:
            event.set()
            return
        banner()
        hours = read(msg=_('How many hours should be left until the wine runs out in a city so that it\'s alerted?'), min=1)
        print(_('It will be alerted when the wine runs out in less than {:d} hours in any city').format(hours))
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nI alert if the wine runs out in less than {:d} hours\n').format(hours)
    setInfoSignal(session, info)
    try:
        do_it(session, hours)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def do_it(session, hours):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    hours : int
    """

    was_alerted = {}
    while True:
        # getIdsOfCities is called on a loop because the amount of cities may change
        ids, cities = getIdsOfCities(session)

        for cityId in cities:
            if cityId not in was_alerted:
                was_alerted[cityId] = False

        for cityId in cities:
            html = session.get(city_url + cityId)
            city = getCity(html)

            # if the city doesn't even have a tavern built, ignore it
            if 'tavern' not in [building['building'] for building in city['position']]:
                continue

            consumption_per_hour = city['consumo']

            # is a wine city
            if cities[cityId]['tradegood'] == '1':
                wine_production = getProductionPerSecond(session, cityId)[1]
                wine_production = wine_production * 60 * 60
                if consumption_per_hour > wine_production:
                    consumption_per_hour -= wine_production
                else:
                    was_alerted[cityId] = False
                    continue

            consumption_per_seg = Decimal(consumption_per_hour) / Decimal(3600)
            wine_available = city['recursos'][1]

            if consumption_per_seg == 0:
                if was_alerted[cityId] is False:
                    msg = _('The city {} is not consuming wine!').format(city['name'])
                    sendToBot(session, msg)
                    was_alerted[cityId] = True
                continue

            seconds_left = Decimal(wine_available) / Decimal(consumption_per_seg)
            if seconds_left < hours*60*60:
                if was_alerted[cityId] is False:
                    time_left = daysHoursMinutes(seconds_left)
                    msg = _('In {}, the wine will run out in {}').format(time_left, city['name'])
                    sendToBot(session, msg)
                    was_alerted[cityId] = True
            else:
                was_alerted[cityId] = False

        time.sleep(20*60)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import time
import json
import math
import random
import gettext
import traceback
import threading
from decimal import *
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.varios import *
from ikabot.helpers.botComm import *
from ikabot.helpers.pedirInfo import *
from ikabot.web.session import normal_get
from ikabot.helpers.planRoutes import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.resources import getAvailableResources
t = gettext.translation('constructionList', localedir, languages=languages, fallback=True)
_ = t.gettext

sendResources = True
expand = True
thread = None

def waitForConstruction(session, city_id):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city_id : int

    Returns
    -------
    city : dict
    """
    while True:

        html = session.get(city_url + city_id)
        city = getCity(html)

        construction_buildings = [building for building in city['position'] if 'completed' in building]
        if len(construction_buildings) == 0:
            break

        construction_building = construction_buildings[0]
        construction_time = construction_building['completed']

        current_time = int(time.time())
        final_time = int(construction_time)
        seconds_to_wait = final_time - current_time

        msg = _('{}: I wait {:d} seconds so that {} gets to the level {:d}').format(city['cityName'], seconds_to_wait, construction_building['name'], construction_building['level'] + 1)
        sendToBotDebug(session, msg, debugON_constructionList)

        wait(seconds_to_wait + 10)

    html = session.get(city_url + city_id)
    city = getCity(html)
    return city


def expandBuilding(session, cityId, building, waitForResources):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    cityId : int
    building : dict
    waitForResources : bool
    """
    current_level = building['level']
    if building['isBusy']:
        current_level += 1
    levels_to_upgrade = building['upgradeTo'] - current_level
    position = building['position']

    time.sleep(random.randint(5, 15))  # to avoid race conditions with sendResourcesNeeded

    for lv in range(levels_to_upgrade):
        city = waitForConstruction(session, cityId)
        building = city['position'][position]

        if building['canUpgrade'] is False and waitForResources is True:
            while building['canUpgrade'] is False:
                time.sleep(60)
                seconds = getMinimumWaitingTime(session)
                html = session.get(city_url + cityId)
                city = getCity(html)
                building = city['position'][position]
                # if no ships are comming, exit no matter if the building can or can't upgrade
                if seconds == 0:
                    break
                wait(seconds + 5)

        if building['canUpgrade'] is False:
            msg = _('City:{}\n').format(city['cityName'])
            msg += _('Building:{}\n').format(building['name'])
            msg += _('The building could not be completed due to lack of resources.\n')
            msg += _('Missed {:d} levels').format(levels_to_upgrade - lv)
            sendToBot(session, msg)
            return

        url = 'action=CityScreen&function=upgradeBuilding&actionRequest={}&cityId={}&position={:d}&level={}&activeTab=tabSendTransporter&backgroundView=city&currentCityId={}&templateView={}&ajax=1'.format(actionRequest, cityId, position, building['level'], cityId, building['building'])
        resp = session.post(url)
        html = session.get(city_url + cityId)
        city = getCity(html)
        building = city['position'][position]
        if building['isBusy'] is False:
            msg = _('{}: The building {} was not extended').format(city['cityName'], building['name'])
            sendToBot(session, msg)
            sendToBot(session, resp)
            return

        msg = _('{}: The building {} is being extended to level {:d}.').format(city['cityName'], building['name'], building['level']+1)
        sendToBotDebug(session, msg, debugON_constructionList)

    msg = _('{}: The building {} finished extending to level: {:d}.').format(city['cityName'], building['name'], building['level']+1)
    sendToBotDebug(session, msg, debugON_constructionList)


def getCostsReducers(city):
    """
    Parameters
    ----------
    city : dict

    Returns
    -------
    reducers_per_material_level : dict[int, int]
    """
    reducers_per_material = [0] * len(materials_names)
    assert len(reducers_per_material) == 5

    for building in city['position']:
        if building['name'] == 'empty':
            continue
        lv = building['level']
        if building['building'] == 'carpentering':
            reducers_per_material[0] = lv
        elif building['building'] == 'vineyard':
            reducers_per_material[1] = lv
        elif building['building'] == 'architect':
            reducers_per_material[2] = lv
        elif building['building'] == 'optician':
            reducers_per_material[3] = lv
        elif building['building'] == 'fireworker':
            reducers_per_material[4] = lv
    return reducers_per_material


def getResourcesNeeded(session, city, building, current_level, final_level):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    building : dict
    current_level : int
    final_level : int

    Returns
    -------
    costs_per_material : dict[int, int]
    """
    # get html with information about buildings
    building_detail_url = 'view=buildingDetail&buildingId=0&helpId=1&backgroundView=city&currentCityId={}&templateView=ikipedia&actionRequest={}&ajax=1'.format(city['id'], actionRequest)
    building_detail_response = session.post(building_detail_url)
    building_detail = json.loads(building_detail_response, strict=False)
    building_html = building_detail[1][1][1]

    # get html with information about buildings costs
    regex_building_detail = r'<div class="(?:selected)? button_building ' + re.escape(building['building']) + r'"\s*onmouseover="\$\(this\)\.addClass\(\'hover\'\);" onmouseout="\$\(this\)\.removeClass\(\'hover\'\);"\s*onclick="ajaxHandlerCall\(\'\?(.*?)\'\);'
    match = re.search(regex_building_detail, building_html)
    building_costs_url = match.group(1)
    building_costs_url += 'backgroundView=city&currentCityId={}&templateView=buildingDetail&actionRequest={}&ajax=1'.format(city['id'], actionRequest)
    building_costs_response = session.post(building_costs_url)
    building_costs = json.loads(building_costs_response, strict=False)
    html_costs = building_costs[1][1][1]

    # if the user has all the resource saving studies, we save that in the session data (one less request)
    sessionData = session.getSessionData()
    if 'reduccion_inv_max' in sessionData:
        costs_reduction = 14
    else:
        # get the studies
        url = 'view=noViewChange&researchType=economy&backgroundView=city&currentCityId={}&templateView=researchAdvisor&actionRequest={}&ajax=1'.format(city['id'], actionRequest)
        rta = session.post(url)
        rta = json.loads(rta, strict=False)
        studies = rta[2][1]['new_js_params']
        studies = json.loads(studies, strict=False)
        studies = studies['currResearchType']

        # look for resource saving studies
        costs_reduction = 0
        for study in studies:
            if studies[study]['liClass'] != 'explored':
                continue
            link = studies[study]['aHref']
            if '2020' in link:
                costs_reduction += 2
            elif '2060' in link:
                costs_reduction += 4
            elif '2100' in link:
                costs_reduction += 8

        # if the user has all the resource saving studies, save that in the session data
        if costs_reduction == 14:
            sessionData['reduccion_inv_max'] = True
            session.setSessionData(sessionData)

    # calculate cost reductions
    costs_reduction /= 100
    costs_reduction = 1 - costs_reduction

    # get buildings that reduce the cost of upgrades
    costs_reductions = getCostsReducers(city)

    # get the type of resources that this upgrade will cost (wood, marble, etc)
    resources_types = re.findall(r'<th class="costs"><img src="skin/resources/icon_(.*?)\.png"/></th>', html_costs)[:-1]

    # get the actual cost of each upgrade
    matches = re.findall(r'<td class="level">\d+</td>(?:\s+<td class="costs">.*?</td>)+', html_costs)

    # calculate the cost of the entire upgrade, taking into account all the possible reductions
    final_costs = [0] * len(materials_names)
    levels_to_upgrade = 0
    for match in matches:
        lv = re.search(r'"level">(\d+)</td>', match).group(1)
        lv = int(lv)

        if lv <= current_level:
            continue
        if lv > final_level:
            break

        levels_to_upgrade += 1

        # get the costs for the current level
        costs = re.findall(r'<td class="costs">([\d,\.]*)</td>', match)
        for i in range(len(costs)):
            resource_type = resources_types[i]
            for j in range(len(materials_names_tec)):
                name = materials_names_tec[j]
                if resource_type == name:
                    resource_index = j
                    break

            # get the cost of the current resource type
            cost = costs[i]
            cost = cost.replace(',', '').replace('.', '')
            cost = 0 if cost == '' else int(cost)

            # calculate all the reductions
            real_cost = Decimal(cost)
            # investigation reduction
            original_cost = Decimal(real_cost) / Decimal(costs_reduction)
            # special building reduction
            real_cost -= Decimal(original_cost) * (Decimal(costs_reductions[resource_index]) / Decimal(100))

            final_costs[resource_index] += math.ceil(real_cost)

    if levels_to_upgrade < final_level - current_level:
        print(_('This building only allows you to expand {:d} more levels').format(levels_to_upgrade))
        msg = _('Expand {:d} levels? [Y/n]:').format(levels_to_upgrade)
        rta = read(msg=msg, values=['Y', 'y', 'N', 'n', ''])
        if rta.lower() == 'n':
            return [-1, -1, -1, -1, -1]

    return final_costs


def sendResourcesNeeded(session, destination_city_id, city_origins, missing_resources):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    destination_city_id : int
    city_origins : dict
    missing_resources : dict[int, int]
    """

    info = _('\nTransport resources to upload building\n')

    try:
        routes = []
        html = session.get(city_url + destination_city_id)
        cityD = getCity(html)
        for i in range(len(materials_names)):
            missing = missing_resources[i]
            if missing <= 0:
                continue

            # send the resources from each origin city
            for cityOrigin in city_origins[i]:
                if missing == 0:
                    break

                available = cityOrigin['recursos'][i]
                send = min(available, missing)
                missing -= send
                toSend = [0] * len(materials_names)
                toSend[i] = send
                route = (cityOrigin, cityD, cityD['islandId'], *toSend)
                routes.append(route)
        executeRoutes(session, routes)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
        # no s.logout() because this is a thread, not a process


def chooseResourceProviders(session, cities_ids, cities, city_id, resource, missing):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    cities_ids : list[int]
    cities : dict[int, dict]
    city_id : int
    resource : int
    missing : int
    """
    global sendResources
    sendResources = True
    global expand
    expand = True

    banner()
    print(_('From what cities obtain {}?').format(materials_names[resource].lower()))

    tradegood_initials = [material_name[0] for material_name in materials_names]
    maxName = max([len(cities[city]['name']) for city in cities if cities[city]['id'] != city_id])

    origin_cities = []
    total_available = 0
    for cityId in cities_ids:
        if cityId == city_id:
            continue

        html = session.get(city_url + cityId)
        city = getCity(html)

        available = city['recursos'][resource]
        if available == 0:
            continue

        # ask the user it this city should provide resources
        tradegood_initial = tradegood_initials[int(cities[cityId]['tradegood'])]
        pad = ' ' * (maxName - len(cities[cityId]['name']))
        msg = '{}{} ({}): {} [Y/n]:'.format(pad, cities[cityId]['name'], tradegood_initial, addThousandSeparator(available))
        choice = read(msg=msg, values=['Y', 'y', 'N', 'n', ''])
        if choice.lower() == 'n':
            continue

        # if so, save the city and calculate the total amount resources to send
        total_available += available
        origin_cities.append(city)
        # if we have enough resources, return
        if total_available >= missing:
            return origin_cities

    # if we reach this part, there are not enough resources to expand the building
    print(_('\nThere are not enough resources.'))

    if len(origin_cities) > 0:
        print(_('\nSend the resources anyway? [Y/n]'))
        choice = read(values=['y', 'Y', 'n', 'N', ''])
        if choice.lower() == 'n':
            sendResources = False

    print(_('\nTry to expand the building anyway? [y/N]'))
    choice = read(values=['y', 'Y', 'n', 'N', ''])
    if choice.lower() == 'n' or choice == '':
        expand = False

    return origin_cities


def sendResourcesMenu(session, city_id, missing):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city_id : int
    missing : list[int, int]
    """
    global thread
    cities_ids, cities = getIdsOfCities(session)
    origins = {}
    # for each missing resource, choose providers
    for resource in range(len(missing)):
        if missing[resource] <= 0:
            continue

        origin_cities = chooseResourceProviders(session, cities_ids, cities, city_id, resource, missing[resource])
        if sendResources is False and expand:
            print(_('\nThe building will be expanded if possible.'))
            enter()
            return
        elif sendResources is False:
            return
        origins[resource] = origin_cities

    if expand:
        print(_('\nThe resources will be sent and the building will be expanded if possible.'))
    else:
        print(_('\nThe resources will be sent.'))

    enter()

    # create a new thread to send the resources
    thread = threading.Thread(target=sendResourcesNeeded, args=(session, city_id, origins, missing,))
    thread.start()


def getBuildingToExpand(session, cityId):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    cityId : int

    Returns
    -------
    building : dict
    """
    html = session.get(city_url + cityId)
    city = getCity(html)

    banner()
    # show the buildings available to expand (ignore empty spaces)
    print(_('Which building do you want to expand?\n'))
    print(_('(0)\t\texit'))
    buildings = [building for building in city['position'] if building['name'] != 'empty']
    for i in range(len(buildings)):
        building = buildings[i]

        level = building['level']
        if level < 10:
            level = ' ' + str(level)
        else:
            level = str(level)
        if building['isBusy']:
            level = level + '+'
        print(_('({:d})\tlv:{}\t{}').format(i+1, level, building['name']))

    selected_building_id = read(min=0, max=len(buildings))
    if selected_building_id == 0:
        return None

    building = buildings[selected_building_id - 1]

    current_level = int(building['level'])
    # if the building is being expanded, add 1 level
    if building['isBusy']:
        current_level += 1

    banner()
    print(_('building:{}').format(building['name']))
    print(_('current level:{}').format(current_level))

    final_level = read(min=current_level, msg=_('increase to level:'))
    building['upgradeTo'] = final_level

    return building


def constructionList(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        global expand
        global sendResources
        expand = True
        sendResources = True

        banner()
        wait_resources = False
        print(_('In which city do you want to expand a building?'))
        city = chooseCity(session)
        cityId = city['id']
        building = getBuildingToExpand(session, cityId)
        if building is None:
            event.set()
            return

        current_level = building['level']
        if building['isBusy']:
            current_level += 1
        final_level = building['upgradeTo']

        # calculate the resources that are needed
        resourcesNeeded = getResourcesNeeded(session, city, building, current_level, final_level)
        if -1 in resourcesNeeded:
            event.set()
            return

        print('\nMaterials needed:')
        for i, name in enumerate(materials_names):
            amount = resourcesNeeded[i]
            if amount == 0:
                continue
            print('- {}: {}'.format(name, addThousandSeparator(amount)))
        print('')

        # calculate the resources that are missing
        missing = [0] * len(materials_names)
        for i in range(len(materials_names)):
            if city['recursos'][i] < resourcesNeeded[i]:
                missing[i] = resourcesNeeded[i] - city['recursos'][i]

        # show missing resources to the user
        if sum(missing) > 0:
            print(_('\nMissing:'))
            for i in range(len(materials_names)):
                if missing[i] == 0:
                    continue
                name = materials_names[i].lower()
                print(_('{} of {}').format(addThousandSeparator(missing[i]), name))
            print('')

            # if the user wants, send the resources from the selected cities
            print(_('Automatically transport resources? [Y/n]'))
            rta = read(values=['y', 'Y', 'n', 'N', ''])
            if rta.lower() == 'n':
                print(_('Proceed anyway? [Y/n]'))
                rta = read(values=['y', 'Y', 'n', 'N', ''])
                if rta.lower() == 'n':
                    event.set()
                    return
            else:
                wait_resources = True
                sendResourcesMenu(session, cityId, missing)
        else:
            print(_('\nYou have enough materials'))
            print(_('Proceed? [Y/n]'))
            rta = read(values=['y', 'Y', 'n', 'N', ''])
            if rta.lower() == 'n':
                event.set()
                return
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nUpgrade building\n')
    info = info + _('City: {}\nBuilding: {}. From {:d}, to {:d}').format(city['cityName'], building['name'], current_level, final_level)

    setInfoSignal(session, info)
    try:
        if expand:
            expandBuilding(session, cityId, building, wait_resources)
        elif thread:
            thread.join()
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import gettext
from decimal import *
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.naval import *
from ikabot.helpers.varios import *
from ikabot.helpers.resources import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.market import getGold

t = gettext.translation('getStatus', localedir, languages=languages, fallback=True)
_ = t.gettext

getcontext().prec = 30


def getStatus(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()
        color_arr = [bcolors.ENDC, bcolors.HEADER, bcolors.STONE, bcolors.BLUE, bcolors.WARNING]

        (ids, __) = getIdsOfCities(session)
        total_resources = [0] * len(materials_names)
        total_production = [0] * len(materials_names)
        total_wine_consumption = 0
        available_ships = 0
        total_ships = 0
        for id in ids:
            session.get('view=city&cityId={}'.format(id), noIndex=True)
            data = session.get("view=updateGlobalData&ajax=1", noIndex=True)
            json_data = json.loads(data, strict=False)
            json_data = json_data[0][1]['headerData']
            if json_data['relatedCity']['owncity'] != 1:
                continue
            wood = Decimal(json_data['resourceProduction'])
            good = Decimal(json_data['tradegoodProduction'])
            typeGood = int(json_data['producedTradegood'])
            total_production[0] += wood * 3600
            total_production[typeGood] += good * 3600
            total_wine_consumption += json_data['wineSpendings']
            total_resources[0] += json_data['currentResources']['resource']
            total_resources[1] += json_data['currentResources']['1']
            total_resources[2] += json_data['currentResources']['2']
            total_resources[3] += json_data['currentResources']['3']
            total_resources[4] += json_data['currentResources']['4']
            available_ships = json_data['freeTransporters']
            total_ships = json_data['maxTransporters']
            total_gold = int(Decimal(json_data['gold']))
            total_gold_production = int(Decimal(json_data['scientistsUpkeep'] + json_data['income'] + json_data['upkeep']))
        print(_('Ships {:d}/{:d}').format(int(available_ships), int(total_ships)))
        print(_("\nTotal:"))
        print('{:>10}'.format(' '), end='|')
        for i in range(len(materials_names)):
            print('{:>12}'.format(materials_names_english[i]), end='|')
        print()
        print('{:>10}'.format('Available'), end='|')
        for i in range(len(materials_names)):
            print('{:>12}'.format(addThousandSeparator(total_resources[i], ' ')), end='|')
        print()
        print('{:>10}'.format('Production'), end='|')
        for i in range(len(materials_names)):
            print('{:>12}'.format(addThousandSeparator(total_production[i], ' ')), end='|')
        print()
        print("Gold : {}, Gold production : {}".format(addThousandSeparator(total_gold, ' '), addThousandSeparator(total_gold_production, ' ')))
        print("Wine consumption : {}".format(addThousandSeparator(total_wine_consumption, ' ')), end='')

        print(_('\nOf which city do you want to see the state?'))
        city = chooseCity(session)
        banner()

        (wood, good, typeGood) = getProductionPerSecond(session, city['id'])
        print('\033[1m{}{}{}'.format(color_arr[int(typeGood)], city['cityName'], color_arr[0]))

        resources = city['recursos']
        storageCapacity = city['storageCapacity']
        color_resources = []
        for i in range(len(materials_names)):
            if resources[i] == storageCapacity:
                color_resources.append(bcolors.RED)
            else:
                color_resources.append(bcolors.ENDC)
        print(_('Storage:'))
        print(addThousandSeparator(storageCapacity))
        print(_('Resources:'))
        for i in range(len(materials_names)):
            print('{} {}{}{} '.format(materials_names[i], color_resources[i], addThousandSeparator(resources[i]), bcolors.ENDC), end='')
        print('')

        print(_('Production:'))
        print('{}:{} {}:{}'.format(materials_names[0], addThousandSeparator(wood*3600), materials_names[typeGood], addThousandSeparator(good*3600)))

        hasTavern = 'tavern' in [building['building'] for building in city['position']]
        if hasTavern:
            consumption_per_hour = city['consumo']
            if consumption_per_hour == 0:
                print(_('{}{}Does not consume wine!{}').format(bcolors.RED, bcolors.BOLD, bcolors.ENDC))
            else:
                if typeGood == 1 and (good*3600) > consumption_per_hour:
                    elapsed_time_run_out = '∞'
                else:
                    consumption_per_second = Decimal(consumption_per_hour) / Decimal(3600)
                    remaining_resources_to_consume = Decimal(resources[1]) / Decimal(consumption_per_second)
                    elapsed_time_run_out = daysHoursMinutes(remaining_resources_to_consume)
                print(_('There is wine for: {}').format(elapsed_time_run_out))

        for building in [building for building in city['position'] if building['name'] != 'empty']:
            if building['isMaxLevel'] is True:
                color = bcolors.BLACK
            elif building['canUpgrade'] is True:
                color = bcolors.GREEN
            else:
                color = bcolors.RED

            level = building['level']
            if level < 10:
                level = ' ' + str(level)
            else:
                level = str(level)
            if building['isBusy'] is True:
                level = level + '+'

            print(_('lv:{}\t{}{}{}').format(level, color, building['name'], bcolors.ENDC))

        enter()
        print('')
        event.set()
    except KeyboardInterrupt:
        event.set()
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import json
import gettext
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *

t = gettext.translation('constructBuilding', localedir, languages=languages, fallback=True)
_ = t.gettext


def constructBuilding(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        print(_('City where to build:'))
        city = chooseCity(session)
        banner()

        # list of free spaces in the selected city
        free_spaces = [buildings for buildings in city['position'] if buildings['building'] == 'empty']

        # get a list of all the posible buildings that can be built
        buildings = []
        # different buildings can be built in different areas
        type_spaces = ['sea', 'land', 'shore', 'wall']
        for type_space in type_spaces:
            free_spaces_of_type = [free_space for free_space in free_spaces if free_space['type'] == type_space]
            if len(free_spaces_of_type) > 0:
                # we take any space in the desired area
                free_space_of_type = free_spaces_of_type[0]
                params = {'view': 'buildingGround', 'cityId': city['id'], 'position': free_space_of_type['position'], 'backgroundView': 'city', 'currentCityId': city['id'], 'actionRequest': actionRequest, 'ajax': '1'}
                buildings_response = session.post(params=params, noIndex=True)
                buildings_response = json.loads(buildings_response, strict=False)[1][1]
                if buildings_response == '':
                    continue
                html = buildings_response[1]
                matches = re.findall(r'<li class="building (.+?)">\s*<div class="buildinginfo">\s*<div title="(.+?)"\s*class="buildingimg .+?"\s*onclick="ajaxHandlerCall\(\'.*?buildingId=(\d+)&', html)
                # add the buildings that can be built in this area
                for match in matches:
                    buildings.append({'building': match[0], 'name': match[1], 'buildingId': match[2], 'type': type_space})

        if len(buildings) == 0:
            print(_('No building can be built.'))
            enter()
            event.set()
            return

        # show list of buildings to the user
        print(_('What building do you want to build?\n'))
        i = 0
        for building in buildings:
            i += 1
            print('({:d}) {}'.format(i, building['name']))
        selected_building_index = read(min=1, max=i)
        banner()

        # show posible positions for the selected building
        building = buildings[selected_building_index - 1]
        print('{}\n'.format(building['name']))
        options = [position_id for position_id in city['position'] if position_id['building'] == 'empty' and position_id['type'] == building['type']]
        if len(options) == 1:
            option = options[0]
        else:
            print(_('In which position do you want to build?\n'))
            i = 0
            for option in options:
                i += 1
                print('({:d}) {}'.format(i, option['position']))
            selected_building_index = read(min=1, max=i)
            option = options[selected_building_index - 1]
            banner()

        # build it
        params = {'action': 'CityScreen', 'function': 'build', 'cityId': city['id'], 'position': option['position'], 'building': building['buildingId'], 'backgroundView': 'city', 'currentCityId': city['id'], 'templateView': 'buildingGround', 'actionRequest': actionRequest, 'ajax': '1'}
        buildings_response = session.post(params=params, noIndex=True)
        msg = json.loads(buildings_response, strict=False)[3][1][0]['text']
        print(msg)
        enter()
        event.set()
    except KeyboardInterrupt:
        event.set()
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import gettext
import ikabot.config as config
from ikabot.helpers.process import run

t = gettext.translation('checkForUpdate', config.localedir, languages=config.languages, fallback=True)
_ = t.gettext


def checkForUpdate():
    upgrade = run('python3 -m pip search ikabot')
    if 'ikabot' not in upgrade:
        return

    upgrade = upgrade.split('\n')
    if len(upgrade) != 3:
        return

    upgrade = upgrade[2]
    match = re.search(r' +.*?: +(.*)', upgrade)
    if match is None:
        return

    new = match.group(1)
    config.update_msg = _('[+] ikabot version {} is available\n').format(new)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import os
import time
import json
import traceback
import threading
import sys
from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.gui import enter
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import daysHoursMinutes
from ikabot.function.vacationMode import activateVacationMode

t = gettext.translation('alertAttacks', localedir, languages=languages, fallback=True)
_ = t.gettext


def alertAttacks(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        if checkTelegramData(session) is False:
            event.set()
            return

        banner()
        default = 20
        minutes = read(msg=_('How often should I search for attacks?(min:3, default: {:d}): ').format(default), min=3, default=default)
        # min_units = read(msg=_('Attacks with less than how many units should be ignored? (default: 0): '), digit=True, default=0)
        print(_('I will check for attacks every {:d} minutes').format(minutes))
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nI check for attacks every {:d} minutes\n').format(minutes)
    setInfoSignal(session, info)
    try:
        do_it(session, minutes)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def respondToAttack(session):
    """
    Parameters
    ---------
    session : ikabot.web.session.Session
    """

    # this allows the user to respond to an attack via telegram
    while True:
        time.sleep(60 * 3)
        responses = getUserResponse(session)
        for response in responses:
            # the response should be in the form of:
            # <pid>:<action number>
            rta = re.search(r'(\d+):?\s*(\d+)', response)
            if rta is None:
                continue

            pid = int(rta.group(1))
            action = int(rta.group(2))

            # if the pid doesn't match, we ignore it
            if pid != os.getpid():
                continue

            # currently just one action is supported
            if action == 1:
                # mv
                activateVacationMode(session)
            else:
                sendToBot(session, _('Invalid command: {:d}').format(action))


def do_it(session, minutes):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    minutes : int
    """

    # this thread lets the user react to an attack once the alert is sent
    thread = threading.Thread(target=respondToAttack, args=(session,))
    thread.start()

    knownAttacks = []
    while True:
        # get the militaryMovements
        html = session.get()
        city_id = re.search(r'currentCityId:\s(\d+),', html).group(1)
        url = 'view=militaryAdvisor&oldView=city&oldBackgroundView=city&backgroundView=city&currentCityId={}&actionRequest={}&ajax=1'.format(city_id, actionRequest)
        movements_response = session.post(url)
        postdata = json.loads(movements_response, strict=False)
        militaryMovements = postdata[1][1][2]['viewScriptParams']['militaryAndFleetMovements']
        timeNow = int(postdata[0][1]['time'])

        currentAttacks = []
        for militaryMovement in [mov for mov in militaryMovements if mov['isHostile']]:
            event_id = militaryMovement['event']['id']
            currentAttacks.append(event_id)
            # if we already alerted this, do nothing
            if event_id not in knownAttacks:
                knownAttacks.append(event_id)

                # get information about the attack
                missionText = militaryMovement['event']['missionText']
                origin = militaryMovement['origin']
                target = militaryMovement['target']
                amountTroops = militaryMovement['army']['amount']
                amountFleets = militaryMovement['fleet']['amount']
                timeLeft = int(militaryMovement['eventTime']) - timeNow

                # send alert
                msg = _('-- ALERT --\n')
                msg += missionText + '\n'
                msg += _('from the city {} of {}\n').format(origin['name'], origin['avatarName'])
                msg += _('a {}\n').format(target['name'])
                msg += _('{} units\n').format(amountTroops)
                msg += _('{} fleet\n').format(amountFleets)
                msg += _('arrival in: {}\n').format(daysHoursMinutes(timeLeft))
                msg += _('If you want to put the account in vacation mode send:\n')
                msg += _('{:d}:1').format(os.getpid())
                sendToBot(session, msg)

        # remove old attacks from knownAttacks
        for event_id in list(knownAttacks):
            if event_id not in currentAttacks:
                knownAttacks.remove(event_id)

        time.sleep(minutes * 60)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import getpass
import os
import gettext
from ikabot import config
from ikabot.config import *

t = gettext.translation('gui', localedir, languages=languages, fallback=True)
_ = t.gettext


def enter():
    """Wait for the user to press Enter
    """
    try:
        if config.has_params:
            return
    except Exception:
        pass
    if isWindows:
        input(_('\n[Enter]'))  # TODO improve this
    else:
        getpass.getpass(_('\n[Enter]'))


def clear():
    """Clears all text on the console
    """
    if isWindows:
        os.system('cls')
    else:
        os.system('clear')


def banner():
    """Clears all text on the console and displays the Ikabot ASCII art banner
    """
    clear()
    bner = """
    `7MMF'  `7MM                       `7MM\"""Yp,                 mm
      MM      MM                         MM    Yb                 MM
      MM      MM  ,MP'   ,6"Yb.          MM    dP    ,pW"Wq.    mmMMmm
      MM      MM ;Y     8)   MM          MM\"""bg.   6W'   `Wb     MM
      MM      MM;Mm      ,pm9MM          MM    `Y   8M     M8     MM
      MM      MM `Mb.   8M   MM          MM    ,9   YA.   ,A9     MM
    .JMML.  .JMML. YA.  `Moo9^Yo.      .JMMmmmd9     `Ybmd9'      `Mbmo
    """
    print('\n{}\n\n{}\n{}'.format(bner, config.infoUser, config.update_msg))


class bcolors:
    HEADER = '\033[95m'
    STONE = '\033[37m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    WARNING = '\033[93m'
    RED = '\033[91m'
    BLACK = '\033[90m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import gettext
import traceback
import sys
from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.gui import enter
from ikabot.helpers.pedirInfo import getIdsOfCities
from ikabot.helpers.varios import wait

t = gettext.translation('loginDaily', localedir, languages=languages, fallback=True)
_ = t.gettext


def loginDaily(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()
        print(_('I will enter every day.'))
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nI enter every day\n')
    setInfoSignal(session, info)
    try:
        do_it(session)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def do_it(session):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    """
    while True:
        (ids, cities) = getIdsOfCities(session)
        for id in ids:
            html = session.post(city_url + str(id))
            if 'class="fountain' in html:
                url = 'action=AvatarAction&function=giveDailyActivityBonus&dailyActivityBonusCitySelect={0}&startPageShown=1&detectedDevice=1&autoLogin=on&cityId={0}&activeTab=multiTab2&backgroundView=city&currentCityId={0}&actionRequest={1}&ajax=1'.format(id, actionRequest)
                session.post(url)
                if 'class="fountain_active' in html:
                    url = 'action=AmbrosiaFountainActions&function=collect&backgroundView=city&currentCityId={0}&templateView=ambrosiaFountain&actionRequest={1}&ajax=1'.format(id, actionRequest)
                    session.post(url)
                break
        wait(24*60*60, 60)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import gettext
import traceback
from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.getJson import getCity
from ikabot.helpers.planRoutes import executeRoutes
from ikabot.helpers.resources import *
from ikabot.helpers.varios import addThousandSeparator
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.gui import banner

t = gettext.translation('distributeResources', localedir, languages=languages, fallback=True)
_ = t.gettext


def distributeResources(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        print(_('What resource do you want to distribute?'))
        print(_('(0) Exit'))
        for i in range(len(materials_names)):
            print('({:d}) {}'.format(i+1, materials_names[i]))
        resource = read(min=0, max=5)
        if resource == 0:
            event.set()  # give main process control before exiting
            return
        resource -= 1

        if resource == 0:
            evenly = True
        else:
            print('\nHow do you want to distribute the resources?')
            print('1) From cities that produce them to cities that do not')
            print('2) Distribute them evenly among all cities')
            type_distribution = read(min=1, max=2)
            evenly = type_distribution == 2

        if evenly:
            routes = distribute_evenly(session, resource)
        else:
            routes = distribute_unevenly(session, resource)

        if routes is None:
            event.set()
            return

        banner()
        print(_('\nThe following shipments will be made:\n'))
        for route in routes:
            print('{} -> {} : {} {}'.format(route[0]['name'], route[1]['name'], route[resource+3], materials_names[resource]))  # displays all routes to be executed in console

        print(_('\nProceed? [Y/n]'))
        rta = read(values=['y', 'Y', 'n', 'N', ''])
        if rta.lower() == 'n':
            event.set()
            return

    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()  # this is where we give back control to main process

    info = _('\nDistribute {}\n').format(materials_names[resource])
    setInfoSignal(session, info)

    try:
        executeRoutes(session, routes)  # plan trips for all the routes
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)  # sends message to telegram bot
    finally:
        session.logout()


def distribute_evenly(session, resource_type):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    resource_type : int
    """
    resourceTotal = 0
    (cityIDs, cities) = getIdsOfCities(session)

    originCities = {}
    destinationCities = {}
    allCities = {}
    for cityID in cityIDs:

        html = session.get(city_url + cityID)  # load html from the get request for that particular city
        city = getCity(html)  # convert the html to a city object

        resourceTotal += city['recursos'][resource_type]  # the cities resources are added to the total
        allCities[cityID] = city  # adds the city to all cities

    # if a city doesn't have enough storage to fit resourceAverage
    # ikabot will send enough resources to fill the store to the max
    # then, resourceAverage will be recalculated
    resourceAverage = resourceTotal // len(allCities)
    while True:

        len_prev = len(destinationCities)
        for cityID in allCities:
            if cityID in destinationCities:
                continue
            freeStorage = allCities[cityID]['freeSpaceForResources'][resource_type]
            storage = allCities[cityID]['storageCapacity']
            if storage < resourceAverage:
                destinationCities[cityID] = freeStorage
                resourceTotal -= storage

        resourceAverage = resourceTotal // (len(allCities) - len(destinationCities))

        if len_prev == len(destinationCities):
            for cityID in allCities:
                if cityID in destinationCities:
                    continue
                if allCities[cityID]['recursos'][resource_type] > resourceAverage:
                    originCities[cityID] = allCities[cityID]['recursos'][resource_type] - resourceAverage
                else:
                    destinationCities[cityID] = resourceAverage - allCities[cityID]['recursos'][resource_type]
            break

    originCities = {k: v for k, v in sorted(originCities.items(), key=lambda item: item[1], reverse=True)}  # sort origin cities in descending order
    destinationCities = {k: v for k, v in sorted(destinationCities.items(), key=lambda item: item[1])}  # sort destination cities in ascending order

    routes = []

    for originCityID in originCities:  # iterate through all origin city ids

        for destinationCityID in destinationCities:  # iterate through all destination city ids
            if originCities[originCityID] == 0 or destinationCities[destinationCityID] == 0:
                continue

            if originCities[originCityID] > destinationCities[destinationCityID]:  # if there's more resources above average in the origin city than resources below average in the destination city (origin city needs to have a surplus and destination city needs to have a deficit of resources for a route to be considered)
                toSend = destinationCities[destinationCityID]  # number of resources to send is the number of resources below average in destination city
            else:
                toSend = originCities[originCityID]  # send the amount of resources above average of the current origin city

            if toSend == 0:
                continue

            toSendArr = [0] * len(materials_names)
            toSendArr[resource_type] = toSend
            route = (allCities[originCityID], allCities[destinationCityID], allCities[destinationCityID]['islandId'], *toSendArr)
            routes.append(route)

            # ROUTE BLOCK
            if originCities[originCityID] > destinationCities[destinationCityID]:
                originCities[originCityID] -= destinationCities[destinationCityID]  # remove the sent amount from the origin city's surplus
                destinationCities[destinationCityID] = 0  # set the amount of resources below average in destination city to 0
            else:
                destinationCities[destinationCityID] -= originCities[originCityID]  # remove the sent amount from the amount of resources below average in current destination city
                originCities[originCityID] = 0  # set the amount of resources above average in origin city to 0

    return routes


def distribute_unevenly(session, resource_type):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    resource_type : int
    """
    total_available_resources_from_all_cities = 0
    (cities_ids, cities) = getIdsOfCities(session)
    origin_cities = {}
    destination_cities = {}
    for destination_city_id in cities_ids:
        is_city_mining_this_resource = cities[destination_city_id]['tradegood'] == str(resource_type)
        if is_city_mining_this_resource:
            html = session.get(city_url + destination_city_id)
            city = getCity(html)
            if resource_type == 1:  # wine
                city['available_amount_of_resource'] = city['recursos'][resource_type] - city['consumo'] - 1
            else:
                city['available_amount_of_resource'] = city['recursos'][resource_type]
            if city['available_amount_of_resource'] < 0:
                city['available_amount_of_resource'] = 0
            total_available_resources_from_all_cities += city['available_amount_of_resource']
            origin_cities[destination_city_id] = city
        else:
            html = session.get(city_url + destination_city_id)
            city = getCity(html)
            city['free_storage_for_resource'] = city['freeSpaceForResources'][resource_type]
            if city['free_storage_for_resource'] > 0:
                destination_cities[destination_city_id] = city

    if total_available_resources_from_all_cities <= 0:
        print(_('\nThere are no resources to send.'))
        enter()
        return None
    if len(destination_cities) == 0:
        print(_('\nThere is no space available to send resources.'))
        enter()
        return None

    remaining_resources_to_be_sent_to_each_city = total_available_resources_from_all_cities // len(destination_cities)
    free_storage_available_per_city = [destination_cities[city]['free_storage_for_resource'] for city in destination_cities]
    total_free_storage_available_in_all_cities = sum(free_storage_available_per_city)
    remaining_resources_to_send = min(total_available_resources_from_all_cities, total_free_storage_available_in_all_cities)
    toSend = {}

    while remaining_resources_to_send > 0:
        len_prev = len(toSend)
        for city_id in destination_cities:
            city = destination_cities[city_id]
            if city_id not in toSend and city['free_storage_for_resource'] < remaining_resources_to_be_sent_to_each_city:
                toSend[city_id] = city['free_storage_for_resource']
                remaining_resources_to_send -= city['free_storage_for_resource']

        if len(toSend) == len_prev:
            for city_id in destination_cities:
                if city_id not in toSend:
                    toSend[city_id] = remaining_resources_to_be_sent_to_each_city
            break

        free_storage_available_per_city = [destination_cities[city]['free_storage_for_resource'] for city in destination_cities if city not in toSend]
        if len(free_storage_available_per_city) == 0:
            break
        total_free_storage_available_in_all_cities = sum(free_storage_available_per_city)
        remaining_resources_to_send = min(remaining_resources_to_send, total_free_storage_available_in_all_cities)
        remaining_resources_to_be_sent_to_each_city = remaining_resources_to_send // len(free_storage_available_per_city)

    routes = []
    for destination_city_id in destination_cities:
        destination_city = destination_cities[destination_city_id]
        island_id = destination_city['islandId']
        missing_resources = toSend[destination_city_id]
        for origin_city_id in origin_cities:
            if missing_resources == 0:
                break

            origin_city = origin_cities[origin_city_id]
            resources_available_in_this_city = origin_city['available_amount_of_resource']
            for route in routes:
                origin = route[0]
                resource = route[resource_type + 3]
                if origin['id'] == origin_city_id:
                    resources_available_in_this_city -= resource

            send_this_round = min(missing_resources, resources_available_in_this_city)
            available = destination_city['free_storage_for_resource']
            if available == 0 or send_this_round == 0:
                continue

            if available < send_this_round:
                missing_resources = 0
                send_this_round = available
            else:
                missing_resources -= send_this_round

            toSendArr = [0] * len(materials_names)
            toSendArr[resource_type] = send_this_round
            route = (origin_city, destination_city, island_id, *toSendArr)

            routes.append(route)

    return routes



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import gettext
import traceback
import sys
from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.gui import enter, enter
from ikabot.helpers.varios import wait
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.pedirInfo import getIslandsIds
from ikabot.helpers.getJson import getIsland
from ikabot.helpers.process import set_child_mode

t = gettext.translation('searchForIslandSpaces', localedir, languages=languages, fallback=True)
_ = t.gettext


def searchForIslandSpaces(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        if checkTelegramData(session) is False:
            event.set()
            return
        banner()
        print('Do you want to search for spaces on your islands or a specific set of islands?')
        print('(0) Exit')
        print('(1) Search all islands I have colonised')
        print('(2) Search a specific set of islands')
        choice = read(min=0, max=2)
        islandList = []
        if choice == 0:
            event.set()
            return
        elif choice == 2:
            banner()
            print('Insert the coordinates of each island you want searched like so: X1:Y1, X2:Y2, X3:Y3...')
            coords_string = read()
            coords_string = coords_string.replace(' ', '')
            coords = coords_string.split(',')
            for coord in coords:
                coord = '&xcoord=' + coord
                coord = coord.replace(':', '&ycoord=')
                html = session.get('view=island' + coord)
                island = getIsland(html)
                islandList.append(island['id'])
        else:
            pass

        banner()
        print('How frequently should the islands be searched in minutes (minimum is 3)?')
        time = read(min=3, digit=True)
        banner()
        print('Do you wish to be notified when a fight breaks out and stops on a city on these islands? (Y|N)')
        fights = read(values=['y', 'Y', 'n', 'N'])
        banner()
        print(_('I will search for changes in the selected islands'))
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nI search for new spaces each hour\n')
    setInfoSignal(session, info)
    try:
        do_it(session, islandList, time, fights)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def do_it(session, islandList, time, fights):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    islandList : list[dict]
        A list containing island objects which should be searched, if an empty list is passed, all the user's colonised islands are searched
    time : int
        The time in minutes between two consecutive seraches
    fights : str
        String that can either be y or n. Indicates whether or not to scan for fight activity on islands
    """

    # this dict will contain all the cities from each island
    # as they where in last scan
    cities_before_per_island = {}

    while True:
        # this is done inside the loop because the user may colonize in a new island
        if islandList != []:
            islandsIds = islandList
        else:
            islandsIds = getIslandsIds(session)
        for islandId in islandsIds:
            html = session.get(island_url + islandId)
            island = getIsland(html)
            # cities in the current island
            cities_now = [city_space for city_space in island['cities'] if city_space['type'] != 'empty']  # loads the islands non empty cities into ciudades

            # if we haven't scaned this island before,
            # save it and do nothing
            if islandId not in cities_before_per_island:
                cities_before_per_island[islandId] = cities_now.copy()
            else:
                cities_before = cities_before_per_island[islandId]

                # someone disappeared
                for city_before in cities_before:
                    if city_before['id'] not in [city_now['id'] for city_now in cities_now]:
                        # we didn't find the city_before in the cities_now
                        msg = _('The city {} of the player {} disappeared in {} {}:{} {}').format(city_before['name'], city_before['Name'], materials_names[int(island['tradegood'])], island['x'], island['y'], island['name'])
                        sendToBot(session, msg)

                    if fights.lower() == 'y':
                        for city_now in cities_now:
                            if city_now['id'] == city_before['id']:
                                if 'infos' in city_now and 'infos' not in city_before and 'armyAction' in city_now['infos'] and city_now['infos']['armyAction'] == 'fight':
                                    msg = _('A fight started in the city {} of the player {} on island {} {}:{} {}').format(city_before['name'], city_before['Name'], materials_names[int(island['tradegood'])], island['x'], island['y'], island['name'])
                                    sendToBot(session, msg)
                                if 'infos' not in city_now and 'infos' in city_before and 'armyAction' in city_before['infos'] and city_before['infos']['armyAction'] == 'fight':
                                    msg = _('A fight stopped in the city {} of the player {} on island {} {}:{} {}').format(city_before['name'], city_before['Name'], materials_names[int(island['tradegood'])], island['x'], island['y'], island['name'])
                                    sendToBot(session, msg)

                # someone colonised
                for city_now in cities_now:
                    if city_now['id'] not in [city_before['id'] for city_before in cities_before]:
                        # we didn't find the city_now in the cities_before
                        msg = _('Player {} created a new city {} in {} {}:{} {}').format(city_now['Name'], city_now['name'], materials_names[int(island['tradegood'])], island['x'], island['y'], island['name'])
                        sendToBot(session, msg)
                        
                cities_before_per_island[islandId] = cities_now.copy() # update cities_before_per_island for the current island

        wait(time * 60)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import os
import json
import random
import gettext
import sys
import ikabot.web.session
from ikabot.config import *
import ikabot.config as config
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import read

t = gettext.translation('botComm', localedir, languages=languages, fallback=True)
_ = t.gettext


def sendToBotDebug(session, msg, debugON):
    """This function will send the ``msg`` argument passed to it as a message to the user on Telegram, only if ``debugOn`` is ``True``
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    msg : str
        a string representing the message to send to the user on Telegram
    debugON : bool
        a boolean indicating whether or not to send the message.
    """
    if debugON:
        sendToBot(session, msg)


def sendToBot(session, msg, Token=False, Photo=None):
    """This function will send the ``msg`` argument passed to it as a message to the user on Telegram
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    msg : str
        a string representing the message to send to the user on Telegram
    Token : bool
        a boolean indicating whether or not to attach the process id, the users server, world and Ikariam username to the message
    Photo : bytes
        a bytes object representing a picture to be sent.
    """
    if checkTelegramData(session) is False:
        return
    if Token is False:
        msg = 'pid:{}\n{}\n{}'.format(os.getpid(), config.infoUser, msg)

    sessionData = session.getSessionData()
    telegram_data = sessionData['shared']['telegram']
    if Photo is None:
        ikabot.web.session.normal_get('https://api.telegram.org/bot{}/sendMessage'.format(telegram_data['botToken']), params={'chat_id': telegram_data['chatId'], 'text': msg})
    else:
        # we need to clear the headers here because telegram doesn't like keep-alive, might as well get rid of all headers
        headers = session.s.headers.copy()
        session.s.headers.clear()
        resp = session.s.post('https://api.telegram.org/bot{}/sendDocument'.format(telegram_data['botToken']), files={'document': ('captcha.png', Photo)}, data={'chat_id': telegram_data['chatId'],'caption': msg})
        session.s.headers = headers


def telegramDataIsValid(session):
    """This function checks whether or not there is any Telegram data stored in the .ikabot file
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    valid : bool
        a boolean indicating whether or not there is any Telegram data stored in the .ikabot file

    """
    sessionData = session.getSessionData()
    try:
        return len(sessionData['shared']['telegram']['botToken']) > 0 and len(sessionData['shared']['telegram']['chatId']) > 0
    except KeyError:
        return False


def getUserResponse(session, fullResponse=False):
    """This function will retrieve a list of messages the user sent to the bot on Telegram.
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    updates : list[str]
        a list containing all the messages the user sent to the bot on Telegram
    """
    # returns messages that the user sends to the telegram bot

    if checkTelegramData(session) is False:
        return []

    sessionData = session.getSessionData()
    telegram_data = sessionData['shared']['telegram']

    try:
        updates = ikabot.web.session.normal_get('https://api.telegram.org/bot{}/getUpdates'.format(telegram_data['botToken'])).text
        updates = json.loads(updates, strict=False)
        if updates['ok'] is False:
            return []
        updates = updates['result']
        # only return messages from the chatId of our user
        if fullResponse:
            return [update['message'] for update in updates if 'message' in update and update['message']['chat']['id'] == int(telegram_data['chatId'])]
        else:
            return [update['message']['text'] for update in updates if 'message' in update and update['message']['chat']['id'] == int(telegram_data['chatId'])]
    except KeyError:
        return []


def checkTelegramData(session):
    """This function doesn't actually check any data itself, that is done by the ``telegramDataIsValid`` function. This function returns ``True`` if there is any Telegram data in the .ikabot file, and if there is none, it will ask the user to input it.
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    valid : bool
        a boolean indicating whether or not there is valid Telegram data in the .ikabot file.
    """
    if telegramDataIsValid(session):
        return True
    else:
        banner()
        print(_('You must provide valid credentials to communicate by telegram.'))
        print(_('You require the token of the bot you are going to use and your chat_id'))
        print(_('For more information about how to obtain them read the readme at https://github.com/physics-sp/ikabot'))
        rta = read(msg=_('Will you provide the credentials now? [y/N]'), values=['y', 'Y', 'n', 'N', ''])
        if rta.lower() != 'y':
            return False
        else:
            return updateTelegramData(session)


def updateTelegramData(session, event=None, stdin_fd=None, predetermined_input=[]):
    """This function asks the user to input the Telegram bot's token and the user's own Telegram chat id. After the user has inputted the neccessary data, this function will generate a random 4 digit number, send it to the user as a Telegram message using the token the user provided. It will then ask the user to input that number as validation.
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    event : multiprocessing.Event
        an event which, when fired, gives back control of the terminal to the main process
    stdin_fd : int
        the standard input file descriptor passed to the function as a means of gaining control of the terminal
    predetermined_input : multiprocessing.managers.SyncManager.list
        a process synced list of predetermined inputs

    Returns
    -------
    valid : bool
        a boolean indicating whether or not the Telegram data has been successfully updated
    """
    if event is not None and stdin_fd is not None:
        sys.stdin = os.fdopen(stdin_fd)  # give process access to terminal
    config.predetermined_input = predetermined_input
    banner()
    print(_('To create your own Telegram Bot, read this: https://core.telegram.org/bots#3-how-do-i-create-a-bot'))
    print(_('Just talk to @botfather in Telegram, send /newbot and then choose the bot\'s name.'))
    print(_('Talk to your new bot and send /start'))
    print(_('Remember to keep the token secret!\n'))
    botToken = read(msg=_('Bot\'s token:'))

    updates = ikabot.web.session.normal_get('https://api.telegram.org/bot{}/getUpdates'.format(botToken)).json()
    if 'ok' not in updates or updates['ok'] is False:
        print(_('invalid telegram bot, try again.'))
        enter()
        if event is not None and stdin_fd is not None:
            event.set()
        return False

    user_ids = []
    users = []
    for update in updates['result']:
        if 'message' in update:
            user = update['message']['from']
            if user['id'] not in user_ids and 'username' in user:
                users.append(user)
                user_ids.append(user['id'])

    if len(users) == 0:
        print(_('make sure your personal Telegram account has a username configured and then send a random message to your bot'))
        enter()
        if event is not None and stdin_fd is not None:
            event.set()
        return False
    elif len(users) == 1:
        resp = read(msg=_('is your username {}? [Y/n]').format(users[0]['username']), default='y', values=['y', 'Y', 'N', 'n'])
        if resp.lower() == 'n':
            print(_('talk to your bot and try again'))
            if event is not None and stdin_fd is not None:
                event.set()
            return False
        else:
            chat_id = users[0]['id']
    else:
        print(_('select your username:'))
        print(_('0) My username is not listed'))
        for i, user in enumerate(users):
            print(_('{:d}) {}').format(i+1, user['username']))
        resp = read(min=0, max=len(users))
        if resp == 0:
            print(_('talk to your bot and try again'))
            if event is not None and stdin_fd is not None:
                event.set()
            return False
        else:
            chat_id = users[resp - 1]['id']

    telegram_data = {}
    telegram_data['telegram'] = {}
    telegram_data['telegram']['botToken'] = botToken.replace(' ', '')
    telegram_data['telegram']['chatId'] = str(chat_id)
    session.setSessionData(telegram_data, shared=True)

    rand = str(random.randint(0, 9999)).zfill(4)
    msg = _('The token is:{}').format(rand)
    sendToBot(session, msg, Token=True)

    rta = read(msg=_('A message was sent by telegram, did you receive it? [Y/n]'), values=['y', 'Y', 'n', 'N', ''])
    if rta.lower() == 'n':
        valid = False
    else:
        recibido = read(msg=_('Enter the received token in telegram:'))
        if rand != recibido:
            print(_('The token is incorrect'))
            valid = False
        else:
            print(_('The token is correct'))
            valid = True

    if valid is False:
        telegram_data['telegram']['botToken'] = ''
        telegram_data['telegram']['chatId'] = ''
        session.setSessionData(telegram_data, shared=True)
        print(_('Check the credentials and re-supply them.'))
    else:
        print(_('The data was saved.'))
    enter()

    if event is not None and stdin_fd is not None:
        event.set()  # give main process control before exiting
    return valid




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import time
import json
import math
import gettext
import traceback
from decimal import *
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.varios import *
from ikabot.helpers.botComm import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.naval import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.getJson import getCity
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.planRoutes import waitForArrival

t = gettext.translation('attackBarbarians', localedir, languages=languages, fallback=True)
_ = t.gettext

getcontext().prec = 30


def choose_island(session):
    idsIslands = getIslandsIds(session)
    islands = []
    for idIsland in idsIslands:
        html = session.get(island_url + idIsland)
        island = getIsland(html)
        islands.append(island)

    if len(islands) == 1:
        return islands[0]

    islands.sort(key=lambda island: island['id'])

    longest_island_name_length = 0
    for island in islands:
        longest_island_name_length = max(len(island['name']), longest_island_name_length)

    def pad(island_name):
        return ' ' * (longest_island_name_length - len(island_name)) + island_name

    print(_('In which island do you want to attack the barbarians?'))
    print(_(' 0) Exit'))
    for i, island in enumerate(islands):
        num = ' ' + str(i+1) if i < 9 else str(i+1)
        if island['barbarians']['destroyed'] == 1:
            warn = _('(currently destroyed)')
        else:
            warn = ''
        print(_('{}) [{}:{}] {} ({}) : barbarians lv: {} ({}) {}').format(num, island['x'], island['y'], pad(island['name']), materials_names[int(island['tradegood'])][0].upper(), island['barbarians']['level'], island['barbarians']['city'], warn))

    index = read(min=0, max=len(islands))
    if index == 0:
        return None
    else:
        return islands[index-1]


def get_barbarians_lv(session, island):
    params = {"view": "barbarianVillage", "destinationIslandId": island['id'], "oldBackgroundView": "city", "cityWorldviewScale": "1", "islandId": island['id'], "backgroundView": "island", "currentIslandId": island['id'], "actionRequest": actionRequest, "ajax": "1"}
    resp = session.post(params=params)
    resp = json.loads(resp, strict=False)

    level = int(resp[2][1]['js_islandBarbarianLevel']['text'])
    gold = int(resp[2][1]['js_islandBarbarianResourcegold']['text'].replace(',', ''))

    resources = [0] * len(materials_names)
    for i in range(len(materials_names)):
        if i == 0:
            resources[i] = int(resp[2][1]['js_islandBarbarianResourceresource']['text'].replace(',', ''))
        else:
            resources[i] = int(resp[2][1]['js_islandBarbarianResourcetradegood{:d}'.format(i)]['text'].replace(',', ''))

    html = resp[1][1][1]
    troops = re.findall(r'<div class="army \w*?">\s*<div class=".*?">(.*?)</div>\s*</div>\s*</td>\s*</tr>\s*<tr>\s*<td class="center">\s*(\d+)', html)

    total_cargo = sum(resources)
    ships = math.ceil(Decimal(total_cargo) / Decimal(500))

    info = {
        'island_id': island['id'],
        'level': level,
        'gold': gold,
        'resources': resources,
        'troops': troops,
        'ships': ships
    }

    return info


def get_units(session, city):
    params = {
        'view': 'cityMilitary',
        'activeTab': 'tabUnits',
        'cityId': city['id'],
        'backgroundView': 'city',
        'currentCityId': city['id'],
        'currentTab': 'multiTab1',
        'actionRequest': actionRequest,
        'ajax': '1'
    }

    resp = session.post(params=params)
    resp = json.loads(resp, strict=False)
    html = resp[1][1][1]
    html = html.split('<div class="fleet')[0]

    unit_id_names = re.findall(r'<div class="army (.*?)">\s*<div class="tooltip">(.*?)<\/div>', html)
    unit_amounts = re.findall(r'<td>(.*?)\s*</td>', html)

    units = {}
    for i in range(len(unit_id_names)):
        amount = int(unit_amounts[i].replace(',', '').replace('-', '0'))
        unit_id = unit_id_names[i][0][1:]
        unit_name = unit_id_names[i][1]
        units[unit_id] = {}
        units[unit_id]['name'] = unit_name
        units[unit_id]['amount'] = amount

    return units


def plan_attack(session, city, babarians_info):
    total_units = get_units(session, city)

    if sum([total_units[unit_id]['amount'] for unit_id in total_units]) == 0:
        print('You don\'t have any troops in this city!')
        enter()
        return None

    plan = []
    total_ships = None
    last = False
    while True:

        banner()

        units_available = {}
        for unit_id in total_units:

            already_sent = sum([p['units'][u] for p in plan for u in p['units'] if u == unit_id])
            if already_sent < total_units[unit_id]['amount']:
                units_available[unit_id] = {}
                units_available[unit_id]['amount'] = total_units[unit_id]['amount'] - already_sent
                units_available[unit_id]['name'] = total_units[unit_id]['name']

        if len(units_available) == 0:
            print(_('No more troops available to send'))
            enter()
            break

        attack_round = {}
        attack_round['units'] = {}
        print(_('Which troops do you want to send?'))
        for unit_id in units_available:
            unit_amount = units_available[unit_id]['amount']
            unit_name = units_available[unit_id]['name']
            amount_to_send = read(msg='{} (max: {}): '.format(unit_name, addThousandSeparator(unit_amount)), max=unit_amount, default=0)
            if amount_to_send > 0:
                attack_round['units'][unit_id] = amount_to_send
        print('')

        attack_round['loot'] = last
        if last:
            attack_round['round'] = len(plan) + 1
        else:
            if len(plan) > 0:
                round_def = len(plan) + 1
                attack_round['round'] = read(msg=_('In which battle round do you want to send them? (min: 2, default: {:d}): ').format(round_def), min=2, default=round_def)
            else:
                attack_round['round'] = 1
        print('')

        if last is False:
            if total_ships is None:
                total_ships = getTotalShips(session)
            max_ships = total_ships - sum([ar['ships'] for ar in plan])
            if max_ships > 0:
                attack_round['ships'] = read(msg=_('How many ships do you want to send in this round? (min: 0, max: {:d}): ').format(max_ships), min=0, max=max_ships)
                print('')
            else:
                attack_round['ships'] = 0

        plan.append(attack_round)

        if last:
            break

        print(_('Do you want to send another round of troops? [y/N]'))
        resp = read(values=['y', 'Y', 'n', 'N'], default='n')
        if resp.lower() != 'y':
            print('')
            print(_('Do you want to select the troops that will be used to collect the remaining resources? (they need to destroy the wall) [y/N]'))
            resp = read(values=['y', 'Y', 'n', 'N'], default='n')
            if resp.lower() != 'y':
                break
            else:
                last = True

    plan.sort(key=lambda ar: ar['round'])
    return plan


def attackBarbarians(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        island = choose_island(session)
        if island is None:
            event.set()
            return

        babarians_info = get_barbarians_lv(session, island)

        banner()
        print(_('The barbarians have:'))
        for name, amount in babarians_info['troops']:
            print(_('{} units of {}').format(amount, name))
        print('')

        banner()
        print(_('From which city do you want to attack?'))
        city = chooseCity(session)

        plan = plan_attack(session, city, babarians_info)
        if plan is None:
            event.set()
            return

        banner()
        print(_('The barbarians in [{}:{}] will be attacked.').format(island['x'], island['y']))
        enter()

    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nI attack the barbarians in [{}:{}]\n').format(island['x'], island['y'])
    setInfoSignal(session, info)
    try:
        do_it(session, island, city, babarians_info, plan)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def get_unit_data(session, city_id, unit_id):
    params_w = {
        'view': 'unitdescription',
        'unitId': unit_id,
        'helpId': 9,
        'subHelpId': 0,
        'backgroundView': 'city',
        'currentCityId': city_id,
        'templateView': 'unitdescription',
        'actionRequest': actionRequest,
        'ajax': 1
    }
    resp = session.post(params=params_w)
    resp = json.loads(resp, strict=False)
    html = resp[1][1][1]

    weight = re.search(r'<li class="weight fifthpos" title=".*?"><span\s*class="accesshint">\'.*?\': </span>(\d+)</li>', html).group(1)
    weight = int(weight)

    speed = re.search(r'(\d+)\s*<br\/>\s*<span class="textLabel">.*?\s*:<\/span>\d+<br\/>\s*<\/div>\s*<div class="clearfloat"><\/div>\s*<div class="weapon">', html).group(1)
    speed = int(speed)

    return {'speed': speed, 'weight': weight}


def city_is_in_island(city, island):
    return city['x'] == island['x'] and city['y'] == island['y']


def get_barbarians_info(session, island_id):
    query = {
        'view': 'barbarianVillage',
        'destinationIslandId': island_id,
        'backgroundView': 'island',
        'currentIslandId': island_id,
        'actionRequest': actionRequest,
        'ajax': 1
    }
    resp = session.post(params=query)
    resp = json.loads(resp, strict=False)
    return resp


def wait_until_can_attack(session, city, island, travel_time=0):
    html = session.get(island_url + island['id'])
    island = getIsland(html)

    if island['barbarians']['underAttack'] == 0 and island['barbarians']['destroyed'] == 0:
        # an attack might be on its way
        wait_for_arrival(session, city, island)
        html = session.get(island_url + island['id'])
        island = getIsland(html)

    if island['barbarians']['underAttack'] == 1:
        # a battle is taking place
        attacks = get_current_attacks(session, city['id'], island['id'])
        attacks_fighting = filter_fighting(attacks)
        eventTimes = [attack['eventTime'] for attack in attacks_fighting]
        if len(eventTimes) > 0:
            wait_time = max(eventTimes)
            wait_time -= time.time()
            wait(wait_time + 5)
        wait_until_can_attack(session, city, island, travel_time)

    if island['barbarians']['destroyed'] == 1:
        # the barbarians are destroyed and can't be attacked
        resp = get_barbarians_info(session, island['id'])
        if 'barbarianCityCooldownTimer' in resp[2][1]:
            wait_time = resp[2][1]['barbarianCityCooldownTimer']['countdown']['enddate']
            wait_time -= time.time()
            wait_time -= travel_time
            wait(wait_time + 5)
        wait_until_can_attack(session, city, island, travel_time)


def get_movements(session, city_id=None):
    if city_id is None:
        city_id = getCurrentCityId(session)
    query = {
        'view': 'militaryAdvisor',
        'oldView': 'updateGlobalData',
        'cityId': city_id,
        'backgroundView': 'city',
        'currentCityId': city_id,
        'templateView': 'militaryAdvisor',
        'actionRequest': actionRequest,
        'ajax': 1
    }

    resp = session.post(params=query)
    resp = json.loads(resp, strict=False)
    movements = resp[1][1][2]['viewScriptParams']['militaryAndFleetMovements']

    return movements


def get_current_attacks(session, city_id, island_id):

    movements = get_movements(session, city_id)
    curr_attacks = []

    for movement in movements:
        if movement['event']['mission'] != 13:
            continue
        if movement['target']['islandId'] != int(island_id):
            continue
        if movement['event']['isReturning'] != 0:
            continue
        if movement['origin']['cityId'] == -1:
            continue

        curr_attacks.append(movement)

    return curr_attacks


def wait_for_arrival(session, city, island):
    attacks = get_current_attacks(session, city['id'], island['id'])
    attacks = filter_loading(attacks) + filter_traveling(attacks)
    eventTimes = [attack['eventTime'] for attack in attacks]

    if len(eventTimes) == 0:
        return

    wait_time = max(eventTimes)
    wait_time -= time.time()
    wait(wait_time + 5)

    wait_for_arrival(session, city, island)


def wait_for_round(session, city, island, travel_time, battle_start, round_number):
    if round_number == 1:
        wait_until_can_attack(session, city, island, travel_time)
    else:
        wait_time = battle_start + (round_number - 2) * 15 * 60
        wait_time -= time.time()
        wait_time -= travel_time
        wait(wait_time + 5)

        if battle_start < time.time():
            html = session.get(island_url + city['id'])
            island = getIsland(html)
            assert island['barbarians']['underAttack'] == 1, "the battle ended before expected"


def calc_travel_time(city, island, speed):
    if city['x'] == island['x'] and city['y'] == island['y']:
        return math.ceil(36000/speed)
    else:
        return math.ceil(1200 * math.sqrt(((city['x'] - island['x']) ** 2) + ((city['y'] - island['y']) ** 2)))


def filter_loading(attacks):
    return [attack for attack in attacks if attack['event']['missionState'] == 1]


def filter_traveling(attacks):
    return [attack for attack in attacks if attack['event']['missionState'] == 2 and attack['event']['canAbort']]


def filter_fighting(attacks):
    return [attack for attack in attacks if attack['event']['missionState'] == 2 and attack['event']['canRetreat']]


def wait_until_attack_is_over(session, city, island):
    html = session.get(island_url + island['id'])
    island = getIsland(html)

    while island['barbarians']['destroyed'] == 0:

        attacks = get_current_attacks(session, city['id'], island['id'])
        # the attack probably failed
        if len(attacks) == 0:
            return

        eventTimes = [attack['eventTime'] for attack in attacks]
        wait_time = min(eventTimes)
        wait_time -= time.time()
        wait(wait_time + 5)

        html = session.get(island_url + island['id'])
        island = getIsland(html)


def load_troops(session, city, island, attack_round, units_data, attack_data, extra_cargo=0):
    ships_needed = Decimal(extra_cargo) / Decimal(500)
    speeds = []
    current_units = get_units(session, city)
    for unit_id in attack_round['units']:
        amount_to_send = min(attack_round['units'][unit_id], current_units[unit_id]['amount'])
        attack_data['cargo_army_{}'.format(unit_id)] = amount_to_send

        if unit_id not in units_data:
            units_data[unit_id] = get_unit_data(session, city['id'], unit_id)

        speeds.append(units_data[unit_id]['speed'])

        if city_is_in_island(city, island) is False:
            weight = units_data[unit_id]['weight']
            ships_needed += Decimal(amount_to_send * weight) / Decimal(500)

    ships_needed = math.ceil(ships_needed)
    speed = min(speeds)
    travel_time = calc_travel_time(city, island, speed)
    return attack_data, ships_needed, travel_time


def loot(session, island, city, units_data, loot_round):
    while True:

        attack_data = {'action': 'transportOperations', 'function': 'attackBarbarianVillage', 'actionRequest': actionRequest, 'islandId': island['id'], 'destinationCityId': 0, 'cargo_army_304_upkeep': 3, 'cargo_army_304': 0, 'cargo_army_315_upkeep': 1, 'cargo_army_315': 0, 'cargo_army_302_upkeep': 4, 'cargo_army_302': 0, 'cargo_army_303_upkeep': 3, 'cargo_army_303': 0, 'cargo_army_312_upkeep': 15, 'cargo_army_312': 0, 'cargo_army_309_upkeep': 45, 'cargo_army_309': 0, 'cargo_army_307_upkeep': 15, 'cargo_army_307': 0, 'cargo_army_306_upkeep': 25, 'cargo_army_306': 0, 'cargo_army_305_upkeep': 30, 'cargo_army_305': 0, 'cargo_army_311_upkeep': 20, 'cargo_army_311': 0, 'cargo_army_310_upkeep': 10, 'cargo_army_310': 0, 'transporter': 0, 'barbarianVillage': 1, 'backgroundView': 'island', 'currentIslandId': island['id'], 'templateView': 'plunder', 'ajax': 1}

        # make sure we have ships on the port
        ships_available = waitForArrival(session)

        # if the barbarians are active again or all the resources were stolen, return
        html = session.get(island_url + island['id'])
        island = getIsland(html)
        destroyed = island['barbarians']['destroyed'] == 1
        resources = get_barbarians_lv(session, island)['resources']
        if destroyed is False or sum(resources) == 0:
            return

        # if we already sent an attack and we still have ships on the port, it was the last one
        attacks = get_current_attacks(session, city['id'], island['id'])
        attacks = filter_loading(attacks) + filter_traveling(attacks)
        if len(attacks) > 0:
            return

        attack_data, ships_needed, travel_time = load_troops(session, city, island, loot_round, units_data, attack_data, sum(resources))
        attack_data['transporter'] = min(ships_available, ships_needed)

        # make sure we have time to send the attack
        time_left = None
        resp = get_barbarians_info(session, island['id'])
        if 'barbarianCityCooldownTimer' in resp[2][1]:
            time_left = resp[2][1]['barbarianCityCooldownTimer']['countdown']['enddate']
            time_left -= time.time()
        if time_left is not None and travel_time > time_left:
            return

        # send attack
        session.post(payloadPost=attack_data)


def do_it(session, island, city, babarians_info, plan):

    units_data = {}

    battle_start = None

    for attack_round in plan:

        # this round is supposed to get the resources
        if attack_round['loot']:
            break

        attack_data = {'action': 'transportOperations', 'function': 'attackBarbarianVillage', 'actionRequest': actionRequest, 'islandId': island['id'], 'destinationCityId': 0, 'cargo_army_304_upkeep': 3, 'cargo_army_304': 0, 'cargo_army_315_upkeep': 1, 'cargo_army_315': 0, 'cargo_army_302_upkeep': 4, 'cargo_army_302': 0, 'cargo_army_303_upkeep': 3, 'cargo_army_303': 0, 'cargo_army_312_upkeep': 15, 'cargo_army_312': 0, 'cargo_army_309_upkeep': 45, 'cargo_army_309': 0, 'cargo_army_307_upkeep': 15, 'cargo_army_307': 0, 'cargo_army_306_upkeep': 25, 'cargo_army_306': 0, 'cargo_army_305_upkeep': 30, 'cargo_army_305': 0, 'cargo_army_311_upkeep': 20, 'cargo_army_311': 0, 'cargo_army_310_upkeep': 10, 'cargo_army_310': 0, 'transporter': 0, 'barbarianVillage': 1, 'backgroundView': 'island', 'currentIslandId': island['id'], 'templateView': 'plunder', 'ajax': 1}

        attack_data, ships_needed, travel_time = load_troops(session, city, island, attack_round, units_data, attack_data)

        try:
            wait_for_round(session, city, island, travel_time, battle_start, attack_round['round'])
        except AssertionError:
            # battle ended before expected
            break

        ships_available = waitForArrival(session)
        while ships_available < ships_needed:
            ships_available = waitForArrival(session)
        ships_available -= ships_needed

        # if the number of available troops changed, the POST request might not work as intended

        attack_data['transporter'] = min(babarians_info['ships'], attack_round['ships'], ships_available)

        # send new round
        session.post(payloadPost=attack_data)

        if attack_round['round'] == 1:
            battle_start = time.time() + travel_time

    wait_until_attack_is_over(session, city, island)

    last_round = plan[-1]
    if last_round['loot']:
        loot(session, city, island, units_data, last_round)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import time
import math
import json
import random
from decimal import *
from ikabot.config import *
from ikabot.helpers.varios import wait
from ikabot.helpers.getJson import getCity
from ikabot.helpers.naval import *


def sendGoods(session, originCityId, destinationCityId, islandId, ships, send):
    """This function will execute one route
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    originCityId : int
        integer representing the ID of the origin city
    destinationCityId : int
        integer representing the ID of the destination city
    islandId : int
        integer representing the ID of the destination city's island
    ships : int
        integer representing the amount of ships needed to execute the route
    send : list
        array of resources to send
    """

    # this can fail if a random request is made in between this two posts
    while True:
        html = session.get()
        city = getCity(html)
        currId = city['id']
        data = {'action': 'header', 'function': 'changeCurrentCity', 'actionRequest': actionRequest, 'oldView': 'city', 'cityId': originCityId, 'backgroundView': 'city', 'currentCityId': currId, 'ajax': '1'}
        session.post(payloadPost=data)

        data = {'action': 'transportOperations', 'function': 'loadTransportersWithFreight', 'destinationCityId': destinationCityId, 'islandId': islandId, 'oldView': '', 'position': '', 'avatar2Name': '', 'city2Name': '', 'type': '', 'activeTab': '', 'transportDisplayPrice': '0', 'premiumTransporter': '0', 'minusPlusValue': '500', 'capacity': '5', 'max_capacity': '5', 'jetPropulsion': '0', 'transporters': ships, 'backgroundView': 'city', 'currentCityId': originCityId, 'templateView': 'transport', 'currentTab': 'tabSendTransporter', 'actionRequest': actionRequest, 'ajax': '1'}

        # add amounts of resources to send
        for i in range(len(send)):
            key = 'cargo_resource' if i == 0 else 'cargo_tradegood{:d}'.format(i)
            data[key] = send[i]

        resp = session.post(payloadPost=data)
        resp = json.loads(resp, strict=False)
        if resp[3][1][0]['type'] == 10:
            break
        elif resp[3][1][0]['type'] == 11:
            wait(getMinimumWaitingTime(session))
        time.sleep(5)


def executeRoutes(session, routes):
    """This function will execute all the routes passed to it, regardless if there are enough ships available to do so
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    routes : list
        a list of tuples, each of which represent a route. A route is defined like so : (originCity,destinationCity,islandId,wood,wine,marble,crystal,sulfur). originCity and destintionCity should be passed as City objects
    """
    for route in routes:
        (origin_city, destination_city, island_id, *toSend) = route
        destination_city_id = destination_city['id']

        while sum(toSend) > 0:
            ships_available = waitForArrival(session)
            storageCapacityInShips = ships_available * 500

            html = session.get(city_url + str(origin_city['id']))
            origin_city = getCity(html)
            html = session.get(city_url + str(destination_city_id))
            destination_city = getCity(html)
            foreign = str(destination_city['id']) != str(destination_city_id)
            if foreign is False:
                storageCapacityInCity = destination_city['freeSpaceForResources']

            send = []
            for i in range(len(toSend)):
                if foreign is False:
                    min_val = min(origin_city['recursos'][i], toSend[i], storageCapacityInShips, storageCapacityInCity[i])
                else:
                    min_val = min(origin_city['recursos'][i], toSend[i], storageCapacityInShips)
                send.append(min_val)
                storageCapacityInShips -= send[i]
                toSend[i] -= send[i]

            resources_to_send = sum(send)
            if resources_to_send == 0:
                # no space available
                # wait an hour and try again
                wait(60 * 60)
                continue

            available_ships = int(math.ceil((Decimal(resources_to_send) / Decimal(500))))
            sendGoods(session, origin_city['id'], destination_city_id, island_id, available_ships, send)


def get_random_wait_time():
    return random.randint(0, 20) * 3


def getMinimumWaitingTime(session):
    """This function returns the time needed to wait for the closest fleet to arrive. If all ships are unavailable, this represents the minimum time needed to wait for any ships to become available. A random waiting time between 0 and 10 seconds is added to the waiting time to avoid race conditions between multiple concurrently running processes.
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    timeToWait : int
        the minimum waiting time for the closest fleet to arrive
    """
    html = session.get()
    idCiudad = re.search(r'currentCityId:\s(\d+),', html).group(1)
    url = 'view=militaryAdvisor&oldView=city&oldBackgroundView=city&backgroundView=city&currentCityId={}&actionRequest={}&ajax=1'.format(idCiudad, actionRequest)
    posted = session.post(url)
    postdata = json.loads(posted, strict=False)
    militaryMovements = postdata[1][1][2]['viewScriptParams']['militaryAndFleetMovements']
    current_time = int(postdata[0][1]['time'])
    delivered_times = []
    for militaryMovement in [mv for mv in militaryMovements if mv['isOwnArmyOrFleet']]:
        remaining_time = int(militaryMovement['eventTime']) - current_time
        delivered_times.append(remaining_time)
    if delivered_times:
        return min(delivered_times) + get_random_wait_time()
    else:
        return 0


def waitForArrival(session):
    """This function will return the number of available ships, and if there aren't any, it will wait for the closest fleet to arrive and then return the number of available ships
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    ships : int
        number of available ships
    """
    available_ships = getAvailableShips(session)
    while available_ships == 0:
        minimum_waiting_time_for_ship = getMinimumWaitingTime(session)
        wait(minimum_waiting_time_for_ship)
        available_ships = getAvailableShips(session)
    return available_ships



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import json
from ikabot.config import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.pedirInfo import getIdsOfCities


def getCommercialCities(session):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session

    Returns
    -------
    commercial_cities : list[dict]
    """
    cities_ids = getIdsOfCities(session)[0]
    commercial_cities = []
    for city_id in cities_ids:
        html = session.get(city_url + city_id)
        city = getCity(html)
        for pos, building in enumerate(city['position']):
            if building['building'] == 'branchOffice':
                city['pos'] = pos
                html = getMarketHtml(session, city)
                positions = re.findall(r'<option.*?>(\d+)</option>', html)
                city['rango'] = int(positions[-1])
                commercial_cities.append(city)
                break
    return commercial_cities


def getMarketHtml(session, city):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    """
    url = 'view=branchOffice&cityId={}&position={:d}&currentCityId={}&backgroundView=city&actionRequest={}&ajax=1'.format(city['id'], city['pos'], city['id'], actionRequest)
    data = session.post(url)
    json_data = json.loads(data, strict=False)
    return json_data[1][1][1]


def storageCapacityOfMarket(html):
    match = re.search(r'var\s*storageCapacity\s*=\s*(\d+);', html)
    if match:
        return int(match.group(1))
    else:
        return 0


def onSellInMarket(html):
    mad, vin, mar, cri, azu = re.findall(r'<input type="text" class="textfield"\s*size="\d+"\s*name=".*?"\s*id=".*?"\s*value="(\d+)"', html)
    return [int(mad), int(vin), int(mar), int(cri), int(azu)]


def getGold(session, city):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    Returns
    -------
    gold : int
    """
    url = 'view=finances&backgroundView=city&currentCityId={}&templateView=finances&actionRequest={}&ajax=1'.format(city['id'], actionRequest)
    data = session.post(url)
    json_data = json.loads(data, strict=False)
    gold = json_data[0][1]['headerData']['gold']
    gold = gold.split('.')[0]
    gold = int(gold)
    gold_production = json_data[0][1]['headerData']['scientistsUpkeep'] + json_data[0][1]['headerData']['income'] + json_data[0][1]['headerData']['upkeep']
    return gold, int(gold_production)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import math
import json
import gettext
import traceback
from decimal import *
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.market import *
from ikabot.helpers.botComm import *
from ikabot.helpers.varios import addThousandSeparator, wait
from ikabot.helpers.pedirInfo import read
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.planRoutes import waitForArrival

t = gettext.translation('sellResources', localedir, languages=languages, fallback=True)
_ = t.gettext


def chooseCommercialCity(commercial_cities):
    """
    Parameters
    ----------
    commercial_cities : list[dict]

    Returns
    -------
    commercial_city : dict
    """
    print(_('In which city do you want to sell resources?\n'))
    for i, city in enumerate(commercial_cities):
        print('({:d}) {}'.format(i + 1, city['name']))
    ind = read(min=1, max=len(commercial_cities))
    return commercial_cities[ind - 1]


def getMarketInfo(session, city):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict

    Returns
    -------
    response : dict
    """
    params = {'view': 'branchOfficeOwnOffers', 'activeTab': 'tab_branchOfficeOwnOffers', 'cityId': city['id'], 'position': city['pos'], 'backgroundView': 'city', 'currentCityId': city['id'], 'templateView': 'branchOfficeOwnOffers', 'currentTab': 'tab_branchOfficeOwnOffers', 'actionRequest': actionRequest, 'ajax': '1'}
    resp = session.post(params=params, noIndex=True)
    return json.loads(resp, strict=False)[1][1][1]


def getOffers(session, my_market_city, resource_type):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    my_market_city : dict
    resource_type : int

    Returns
    -------
    offers : list
    """
    if resource_type == 0:
        resource_type = 'resource'
    else:
        resource_type = str(resource_type)
    data = {'cityId': my_market_city['id'], 'position': my_market_city['pos'], 'view': 'branchOffice', 'activeTab': 'bargain', 'type': '333', 'searchResource': resource_type, 'range': my_market_city['rango'], 'backgroundView': 'city', 'currentCityId': my_market_city['id'], 'templateView': 'branchOffice', 'currentTab': 'bargain', 'actionRequest': actionRequest, 'ajax': '1'}
    resp = session.post(payloadPost=data)
    html = json.loads(resp, strict=False)[1][1][1]
    return re.findall(r'<td class=".*?">(.*?)<br/>\((.*?)\)\s*</td>\s*<td>(.*?)</td>\s*<td><img src=".*?"\s*alt=".*?"\s*title=".*?"/></td>\s*<td style="white-space:nowrap;">(\d+)\s*<img src=".*?"\s*class=".*?"/>.*?</td>\s*<td>(\d+)</td>\s*<td><a onclick="ajaxHandlerCall\(this\.href\);return false;"\s*href="\?view=takeOffer&destinationCityId=(\d+)&', html)


def sellToOffers(session, city_to_buy_from, resource_type, event):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city_to_buy_from : dict
    resource_type : int
    event : multiprocessing.Event
    """
    banner()

    offers = getOffers(session, city_to_buy_from, resource_type)

    if len(offers) == 0:
        print(_('No offers available.'))
        enter()
        event.set()
        return

    print(_('Which offers do you want to sell to?\n'))

    chosen_offers = []
    total_amount = 0
    profit = 0
    for offer in offers:
        cityname, username, amount, price, dist, destination_city_id = offer
        cityname = cityname.strip()
        amount = amount.replace(',', '').replace('.', '')
        amount = int(amount)
        price = int(price)
        msg = _('{} ({}): {} at {:d} each ({} in total) [Y/n]').format(cityname, username, addThousandSeparator(amount), price, addThousandSeparator(price*amount))
        rta = read(msg=msg, values=['y', 'Y', 'n', 'N', ''])
        if rta.lower() == 'n':
            continue
        chosen_offers.append(offer)
        total_amount += amount
        profit += amount * price

    if len(chosen_offers) == 0:
        event.set()
        return

    available = city_to_buy_from['recursos'][resource_type]
    amount_to_sell = min(available, total_amount)

    banner()
    print(_('\nHow much do you want to sell? [max = {}]').format(addThousandSeparator(amount_to_sell)))
    amount_to_sell = read(min=0, max=amount_to_sell)
    if amount_to_sell == 0:
        event.set()
        return

    left_to_sell = amount_to_sell
    profit = 0
    for offer in chosen_offers:
        cityname, username, amount, price, dist, destination_city_id = offer
        cityname = cityname.strip()
        amount = amount.replace(',', '').replace('.', '')
        amount = int(amount)
        price = int(price)
        sell = min(amount, left_to_sell)
        left_to_sell -= sell
        profit += sell * price
    print(_('\nSell {} of {} for a total of {}? [Y/n]').format(addThousandSeparator(amount_to_sell), materials_names[resource_type], addThousandSeparator(profit)))
    rta = read(values=['y', 'Y', 'n', 'N', ''])
    if rta.lower() == 'n':
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nI sell {} of {} in {}\n').format(addThousandSeparator(amount_to_sell), materials_names[resource_type], city_to_buy_from['name'])
    setInfoSignal(session, info)
    try:
        do_it1(session, amount_to_sell, chosen_offers, resource_type, city_to_buy_from)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def createOffer(session, my_offering_market_city, resource_type, event):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    my_offering_market_city : dict
    resource_type : int
    event : multiprocessing.Event
    """
    banner()

    html = getMarketInfo(session, my_offering_market_city)
    sell_market_capacity = storageCapacityOfMarket(html)
    total_available_amount_of_resource = my_offering_market_city['recursos'][resource_type]

    print(_('How much do you want to sell? [max = {}]').format(addThousandSeparator(total_available_amount_of_resource)))
    amount_to_sell = read(min=0, max=total_available_amount_of_resource)
    if amount_to_sell == 0:
        event.set()
        return

    price_max, price_min = re.findall(r'\'upper\': (\d+),\s*\'lower\': (\d+)', html)[resource_type]
    price_max = int(price_max)
    price_min = int(price_min)
    print(_('\nAt what price? [min = {:d}, max = {:d}]').format(price_min, price_max))
    price = read(min=price_min, max=price_max)

    print(_('\nI will sell {} of {} at {}: {}').format(addThousandSeparator(amount_to_sell), materials_names[resource_type], addThousandSeparator(price), addThousandSeparator(price * amount_to_sell)))
    print(_('\nProceed? [Y/n]'))
    rta = read(values=['y', 'Y', 'n', 'N', ''])
    if rta.lower() == 'n':
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nI sell {} of {} in {}\n').format(addThousandSeparator(amount_to_sell), materials_names[resource_type], my_offering_market_city['name'])
    setInfoSignal(session, info)
    try:
        do_it2(session, amount_to_sell, price, resource_type, sell_market_capacity, my_offering_market_city)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def sellResources(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        commercial_cities = getCommercialCities(session)
        if len(commercial_cities) == 0:
            print(_('There is no store built'))
            enter()
            event.set()
            return

        if len(commercial_cities) == 1:
            city = commercial_cities[0]
        else:
            city = chooseCommercialCity(commercial_cities)
            banner()

        print(_('What resource do you want to sell?'))
        for index, material_name in enumerate(materials_names):
            print('({:d}) {}'.format(index+1, material_name))
        selected_material = read(min=1, max=len(materials_names))
        resource = selected_material - 1
        banner()

        print(_('Do you want to sell to existing offers (1) or do you want to make your own offer (2)?'))
        selected = read(min=1, max=2)
        [sellToOffers, createOffer][selected - 1](session, city, resource, event)
    except KeyboardInterrupt:
        event.set()
        return


def do_it1(session, left_to_sell, offers, resource_type, city_to_buy_from):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    left_to_sell : int
    offers : list[dict]
    resource_type : int
    city_to_buy_from : dict
    """
    for offer in offers:
        cityname, username, amount, precio, dist, destination_city_id = offer
        cityname = cityname.strip()
        amount_to_buy = amount.replace(',', '').replace('.', '')
        amount_to_buy = int(amount_to_buy)
        while True:
            amount_to_sell = min(amount_to_buy, left_to_sell)
            ships_available = waitForArrival(session)
            ships_needed = math.ceil((Decimal(amount_to_sell) / Decimal(500)))
            ships_used = min(ships_available, ships_needed)
            if ships_needed > ships_used:
                amount_to_sell = ships_used * 500
            left_to_sell -= amount_to_sell
            amount_to_buy -= amount_to_sell

            data = {'action': 'transportOperations', 'function': 'sellGoodsAtAnotherBranchOffice', 'cityId': city_to_buy_from['id'], 'destinationCityId': destination_city_id, 'oldView': 'branchOffice', 'position': city_to_buy_from['pos'], 'avatar2Name': username, 'city2Name': cityname, 'type': '333', 'activeTab': 'bargain', 'transportDisplayPrice': '0', 'premiumTransporter': '0', 'capacity': '5', 'max_capacity': '5', 'jetPropulsion': '0', 'transporters': str(ships_used), 'backgroundView': 'city', 'currentCityId': city_to_buy_from['id'], 'templateView': 'takeOffer', 'currentTab': 'bargain', 'actionRequest': actionRequest, 'ajax': '1'}
            if resource_type == 0:
                data['cargo_resource'] = amount_to_sell
                data['resourcePrice'] = precio
            else:
                data['tradegood{:d}Price'.format(resource_type)] = precio
                data['cargo_tradegood{:d}'.format(resource_type)] = amount_to_sell

            session.get(city_url + city_to_buy_from['id'], noIndex=True)
            session.post(payloadPost=data)

            if left_to_sell == 0:
                return
            if amount_to_buy == 0:
                break


def do_it2(session, amount_to_sell, price, resource_type, sell_market_capacity, city):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    amount_to_sell : int
    price : int
    resource_type : int
    sell_market_capacity : int
    city : dict
    """
    initial_amount_to_sell = amount_to_sell
    html = getMarketInfo(session, city)
    previous_on_sell = onSellInMarket(html)[resource_type]
    while True:
        html = getMarketInfo(session, city)
        currently_on_sell = onSellInMarket(html)[resource_type]
        # if there is space in the store
        if currently_on_sell < storageCapacityOfMarket(html):
            # add our new offer to the free space
            free_space = sell_market_capacity - currently_on_sell
            offer = min(amount_to_sell, free_space)
            amount_to_sell -= offer
            new_offer = currently_on_sell + offer

            payloadPost = {'cityId': city['id'], 'position': city['pos'], 'action': 'CityScreen', 'function': 'updateOffers', 'resourceTradeType': '444', 'resource': '0', 'resourcePrice': '10', 'tradegood1TradeType': '444', 'tradegood1': '0', 'tradegood1Price': '11', 'tradegood2TradeType': '444', 'tradegood2': '0', 'tradegood2Price': '12', 'tradegood3TradeType': '444', 'tradegood3': '0', 'tradegood3Price': '17', 'tradegood4TradeType': '444', 'tradegood4': '0', 'tradegood4Price': '5', 'backgroundView': 'city', 'currentCityId': city['id'], 'templateView': 'branchOfficeOwnOffers', 'currentTab': 'tab_branchOfficeOwnOffers', 'actionRequest': actionRequest, 'ajax': '1'}
            if resource_type == 0:
                payloadPost['resource'] = new_offer
                payloadPost['resourcePrice'] = price
            else:
                payloadPost['tradegood{:d}'.format(resource_type)] = new_offer
                payloadPost['tradegood{:d}Price'.format(resource_type)] = price
            session.post(payloadPost=payloadPost)

            # if we don't have any more to add to the offer, leave the loop
            if amount_to_sell == 0:
                break

        # sleep for 2 hours
        wait(60 * 60 * 2)

    # wait until the last of our offer is actualy bought, and let the user know
    while True:
        html = getMarketInfo(session, city)
        currently_on_sell = onSellInMarket(html)[resource_type]
        if currently_on_sell <= previous_on_sell:
            msg = _('{} of {} was sold at {:d}').format(addThousandSeparator(initial_amount_to_sell), materials_names[resource_type], price)
            sendToBot(session, msg)
            return

        # sleep for 2 hours
        wait(60 * 60 * 2)
